 /****************************************************************************
 *  pptpcomm.c    : $Revision:   1.43  $
 *  Modified      : $Date:   May 13 1996 14:54:52  $
 *  By            : $Author:   tmortsol  $, U.S. Robotics Access Corp.
 *****************************************************************************
 *          Copyright (c) $Date:   May 13 1996 14:54:52  $,                  *
 *            U.S. Robotics Access Corp., Microsoft Corp.                    *
 *                       All Rights Reserved                                 *
 *****************************************************************************
 *                          "PPTP MODULE"                                    *
 * This software may be used to implement the POINT to POINT TUNNELING       *
 * PROTOCOL as defined by the PPTP forum.  The PPTP forum consists of:       *
 * U.S. Robotics, Microsoft, Ascend Communications, 3Com/Primary Access,     *
 * Telematics.                                                               *
 *****************************************************************************
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain this entire header
 *    including the above copyright notice, this list of conditions and
 *    the following disclaimer.
 * 2. The name of U.S. Robotics may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission of U.S. Robotics.
 *
 * THIS SOFTWARE IS PROVIDED BY U.S. ROBOTICS ACESS CORP. ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL U.S. ROBOTICS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/
/*END_HEADER*/

/*
 * pptpcomm.c
 *
 * This file contains the common PPTP protocol engines for the server
 * and client implementations of the PPTP protocol.
 * These routines interface with client and server via API libraries.
 *
 * <Need to talk about network, client, and server APIs here.  This is
 *  still not complete so it will be filled in later when how this is
 *  done becomes more specific>
 *
 * The API is divided into two areas:
 *
 *  - a set of routines referred to as requests that are required by the
 *    PPTP protocol engine to cause specific events to occur.
 *    These routines must be provided by API libraries on the target.
 *
 *  - a set of routines referred to as indications that are provided
 *    by the PPTP protocol engine that must be called by API libraries
 *    on the target when certain events occur.
 *
 * The API routines are generalized so they can work on both a Windows
 * NT client and our device chassis.  It should be noted that these
 * routines may be general enough to port to other environments but this
 * is not guaranteed.
 *
 * Author: Tim Mortsolf, U.S. Robotics Access Corp.
 */

#include "pptpcomm.h"
#include "pptpdata.h"
#include "pptpnapi.h"
#include "pptpfunc.h"

#define STATIC

#if !defined(PPTP_REDUCE_STATIC_STORAGE)

PptpConfig *currentPptpConfig;
PptpCallLineDescriptor *callLineDescriptorTable;
PptpControlChannelDescriptor *controlChannelDescriptorTable;
PptpControlChannelSeekDescriptor *controlChannelSeekDescriptorTable;

STATIC PptpEndpointHandle listenControlEndpointHandle;
STATIC PptpEndpointHandle dataChannelEndpointHandle;

#if defined(PPTP_USE_STRONG_LOCKS)
PptpLock g_PptpLockStrong;
#else
PptpLock g_PptpLockGlobal;
#endif /* defined(PPTP_USE_STRONG_LOCK) */

PptpExternalTimerId externalTimerId;
PptpInternalTimer *internalTimerHead;
PptpInternalTimer *internalTimerTail;

#else

#define listenControlEndpointHandle PPTP_GP(listenControlEndpointHandle)
#define dataChannelEndpointHandle PPTP_GP(dataChannelEndpointHandle)

#define externalTimerId PPTP_GP(externalTimerId)
#define internalTimerHead PPTP_GP(internalTimerHead)
#define internalTimerTail PPTP_GP(internalTimerTail)

PptpGlobals *PptpGp;

#endif /* !defined(PPTP_REDUCE_STATIC_STORAGE) */

STATIC PptpReturnCode PptpContinueControlChannelSeekDescriptorSearch(
	PptpControlChannelSeekDescriptor *controlChannelSeekDescriptorp
	);
STATIC PptpReturnCode PptpAbortControlChannelSeekDescriptorSearch(
	PptpControlChannelSeekDescriptor *controlChannelSeekDescriptorp
	);
STATIC PptpReturnCode PptpContinuePendingCallLineDescriptors(
	PptpControlChannelDescriptor	*controlChannelDescriptorp
	);
STATIC PptpReturnCode PptpContinuePendingControlChannelSeekDescriptors(
	PptpControlChannelDescriptor	*controlChannelDescriptorp
	);
STATIC PptpReturnCode PptpAttemptControlChannelConnection(
	PptpControlChannelDescriptor	*controlChannelDescriptorp
	);
STATIC PptpReturnCode PptpCheckControlChannelSeekDescriptorUsage(
	PptpControlChannelSeekDescriptor	*controlChannelSeekDescriptorp
	);
STATIC PptpReturnCode PptpCheckControlChannelDescriptorUsage(
	PptpControlChannelDescriptor	*controlChannelDescriptorp
	);
STATIC PptpReturnCode PptpControlChannelRemotelyDisconnected(
	PptpControlChannelDescriptor	*controlChannelDescriptorp
	);
STATIC PptpReturnCode PptpShutdownControlChannelDescriptor(
	PptpControlChannelDescriptor	*controlChannelDescriptorp
	);
STATIC PptpReturnCode PptpControlChannelAddressMatch(
	PptpRemoteNetConfigEntry	*firstRemoteNetConfigEntry,
	PptpRemoteNetConfigEntry	*secondRemoteNetConfigEntry
	);
STATIC VOID PptpSetControlChannelAddress(
	PptpRemoteNetConfigEntry	*destRemoteNetConfigEntry,
	PptpRemoteNetConfigEntry	*srcRemoteNetConfigEntry
	);
STATIC PptpReturnCode PptpProtocolControlInput(
	PptpControlChannelDescriptor	*controlChannelDescriptorp,
	PptpControlPacket		*inputPacketp
	);
STATIC VOID PptpServicePendingPacketList(
	PptpCallLineDescriptor		*callLineDescriptorp
	);
STATIC VOID PptpAckRemoteDataChannel(
	PptpCallLineDescriptor		*callLineDescriptorp
	);
STATIC VOID PptpDataChannelStaticTimeoutCallback(
	PptpCallLineDescriptor		*callLineDescriptorp
	);
STATIC PptpSeqNumber PptpSeqNumberDifference(
	PptpSeqNumber			baseSeqNumber,
	PptpSeqNumber			otherSeqNumber
	);
STATIC PptpReturnCode PptpRescheduleControlChannelInactivityTimer(
	PptpControlChannelDescriptor	*controlChannelDescriptorp
	);
STATIC VOID PptpInactivityTimerCallback(
	PptpControlChannelDescriptor	*controlChannelDescriptorp
	);
STATIC VOID PptpEchoResponseTimerCallback(
	PptpControlChannelDescriptor	*controlChannelDescriptorp
	);
STATIC VOID PptpTerminateInactivityIdleProcessing(
	PptpControlChannelDescriptor	*controlChannelDescriptorp
	);
STATIC PptpReturnCode PptpGetInternalTime(
	PptpMsec			*milliSecondsp
	);
STATIC VOID PptpExternalTimerCallback(
	PptpCallbackArgument		callbackArgument
	);
STATIC PptpReturnCode PptpAssignNewControlChannelDescriptor(
	PptpControlChannelDescriptor	**controlChannelDescriptorpp
	);
STATIC VOID PptpTraceControlPacket(
	PptpControlPacket		*inputPacketp,
	BOOL				inputPacket
	);
STATIC VOID PptpTraceDataPacket(
	PptpCallLineNumber		callLineNumber,
	PptpSeqNumber			seqNumber,
	PptpAckNumber			ackNumber,
	UCHAR				*data,
	ULONG				dataLength,
	BOOL				inputPacket
	);

STATIC UCHAR *PptpAddressToString(
	PptpSockaddrIn			*address,
	UCHAR				*buffer
	);

/*
 * Define macros for logging PPTP control and data packet information
 */

//
// dummied out for now!
//
#define PPTP_CONTROL_PACKET_LOG(args)
#define PPTP_DATA_PACKET_HEADER_LOG(args)
#define PPTP_DATA_PACKET_DATA_LOG(data, datalength)

/*
 * PptpSetCurrentConfig()
 *
 * This routine is called to set the contents of the current PPTP
 * configuration table.  This routine must be called before any
 * other PPTP routines
 */

PptpReturnCode PptpSetCurrentConfig(
	PptpConfig *pPptpConfig
	)
{
	ULONG	NumberOfPorts, i;

  DBG_FUNC("PptpSetCurrentConfig", PPTP_PROTOCOL);

  DBG_ENTER();

#if defined(PPTP_REDUCE_STATIC_STORAGE)
  memset(PptpGp, 0, sizeof(*PptpGp));
#endif /* deifned(PPTP_REDUCE_STATIC_STORAGE) */

	if ((currentPptpConfig = (PptpConfig*)ExAllocatePoolWithTag(NonPagedPool,
		                                  sizeof(PptpConfig),
		                                  PPTP_MEMORYPOOL_TAG)) == NULL) {

		PptpPrintf("USRPPTP.SYS: Failed to allocate memory for PptpConfig!\n");
		return PPTP_FAILURE;
	}

	RtlZeroMemory(currentPptpConfig, sizeof(PptpConfig));

	NumberOfPorts = pPptpConfig->numberLineDevices;

	if ((callLineDescriptorTable =
		(PptpCallLineDescriptor*)ExAllocatePoolWithTag(NonPagedPool,
		                           sizeof(PptpCallLineDescriptor) * NumberOfPorts,
								   PPTP_MEMORYPOOL_TAG)) == NULL) {

		ExFreePool(currentPptpConfig);
		
		PptpPrintf("USRPPTP.SYS: Failed to allocate memory for callLineDescriptorTable!\n");
		return PPTP_FAILURE;
	}

	RtlZeroMemory(callLineDescriptorTable,
	sizeof(PptpCallLineDescriptor) * NumberOfPorts);

	if ((controlChannelDescriptorTable =
		(PptpControlChannelDescriptor*)ExAllocatePoolWithTag(NonPagedPool,
		sizeof(PptpControlChannelDescriptor) * NumberOfPorts,
		PPTP_MEMORYPOOL_TAG)) == NULL) {
		
		ExFreePool(callLineDescriptorTable);
		ExFreePool(currentPptpConfig);
		PptpPrintf("USRPPTP.SYS: Failed to allocate memory for controlChannelDescriptorTable!\n");
		return PPTP_FAILURE;
	}

	RtlZeroMemory(controlChannelDescriptorTable,
	sizeof(PptpControlChannelDescriptor) * NumberOfPorts);

	for (i = 0; i < NumberOfPorts; i++) {
		PptpControlChannelDescriptor	*controlChannelDescriptorp =
		&controlChannelDescriptorTable[i];
		controlChannelDescriptorp->controlChannelIndex = i;
	}

	if ((controlChannelSeekDescriptorTable =
		(PptpControlChannelSeekDescriptor*)ExAllocatePoolWithTag(NonPagedPool,
		sizeof(PptpControlChannelSeekDescriptor) * PPTP_MAX_CONTROL_CHANNEL_SEEK_NUMBER,
		PPTP_MEMORYPOOL_TAG)) == NULL) {
		
		ExFreePool(controlChannelDescriptorTable);
		ExFreePool(callLineDescriptorTable);
		ExFreePool(currentPptpConfig);
		PptpPrintf("USRPPTP.SYS: Failed to allocate memory for controlChannelSeekDescriptorTable!\n");
		return PPTP_FAILURE;
	}

	RtlZeroMemory(controlChannelSeekDescriptorTable,
	sizeof(PptpControlChannelSeekDescriptor) * PPTP_MAX_CONTROL_CHANNEL_SEEK_NUMBER);

  /*
   * Copy the contents of the pointer to the network configuration table
   * into NetConfig.
   */

  memcpy(currentPptpConfig, pPptpConfig, sizeof(PptpConfig));

  /*
   * Copy the default remote control control address list into the
   * default control channel seek descriptor
   */

  memcpy(&controlChannelSeekDescriptorTable[0].addressList,
	 &currentPptpConfig->netConfig.defaultRemoteAddressList,
	 sizeof(PptpAddressList));

  DBG_LEAVE();
  return PPTP_SUCCESS;
}


/*
 * PptpInitializeLocks
 *
 * This routine is called to initialize the PPTP synchronization locks.
 * This routine must be called before any other routines that may use
 * the PPTP spinlocks.  Typically, this is done before driver initialization
 */

PptpReturnCode PptpInitializeLocks(
	VOID
	)
{

  /*
   * Initialize the PPTP synchronization locks
   */

#if defined(PPTP_USE_STRONG_LOCKS)
  PPTP_LOCK_INIT(g_PptpLockStrong);
#elif defined(PPTP_USE_STANDARD_LOCKS)
  PPTP_LOCK_INIT(g_PptpLockGlobal);
#endif /* defined(PPTP_USE_STRONG_LOCKS) */

  return PPTP_SUCCESS;
}


/*
 * PptpInit()
 *
 * This routine is called to startup the PPTP protocol engine.
 * It currently starts listening on the control channel TCP port for
 * incoming connections and obtains a GRE endpoint handle.
 */

PptpReturnCode PptpInit(
	VOID
	)
{

  DBG_FUNC("PptpInit", PPTP_PROTOCOL);

  DBG_ENTER();

  /*
   * Start listening for incoming connections on the control
   * channel.
   */

  if (PptpNetControlListenReq(currentPptpConfig->netConfig.localNetConfig.
			      localControlChannelPortNumber,
			      &listenControlEndpointHandle) !=
      PPTP_SUCCESS) {
    ASSERT(!assertOnFatalError);
    DBG_LEAVE();
    return PPTP_FAILURE;
  }

  /*
   * Register for a GRE Handle
   */

  if (PptpNetGREGetHandle(currentPptpConfig->netConfig.localNetConfig.
			  localProtocolNumber,
			  &dataChannelEndpointHandle) != PPTP_SUCCESS) {
    ASSERT(!assertOnFatalError);
    DBG_LEAVE();
    return PPTP_FAILURE;
  }

  /*
   * Set the initial external timer id to PPTP_TIMER_NOT_IN_USE
   */

  externalTimerId = PPTP_TIMER_NOT_IN_USE;

  if (PptpCreateExternalTimerReq(PptpExternalTimerCallback,
	  NULL,
	  250,
	  &externalTimerId) != PPTP_SUCCESS) {
	  ASSERT(!assertOnFatalError);
  }

  DBG_LEAVE();
  return PPTP_SUCCESS;

}


/*
 * PptpNetControlConnectEstablishedInd()
 *
 * This routine is called to indicate that a requested control connection
 * has been established and accepted by the remote endpoint.
 */

PptpReturnCode PptpNetControlConnectEstablishedInd(
	PptpEndpointHandle	endpointHandle
	)
{
  PptpControlChannelDescriptor *controlChannelDescriptorp;
  PptpStartSessionRequestMsg startSessionRequest;

  DBG_FUNC("PptpNetControlConnectEstablishedInd", PPTP_PROTOCOL);

  DBG_ENTER();

  /*
   * Lookup the control channel number for this endpoint handle
   */

  if (PptpLookupControlChannelDescriptorByHandle(endpointHandle,
						 &controlChannelDescriptorp) !=
      PPTP_SUCCESS) {
    ASSERT(!assertOnFatalError);
    DBG_LEAVE();
    return PPTP_FAILURE;
  }

  /*
   * Verify that the control channel connection was locally initiated
   */

  if (!controlChannelDescriptorp->locallyInitiatedConnection) {
    ASSERT(!assertOnFatalError);
    DBG_LEAVE();
    return PPTP_FAILURE;
  }

  /*
   * Verify that the control channel descriptor is in the proper
   * state to receive this event
   */

  switch (controlChannelDescriptorp->controlChannelState) {

  case PPTP_CONTROL_CHANNEL_CONNECTING:
    break;

  case PPTP_CONTROL_CHANNEL_DISCONNECTING:

    /*
     * It is possible to get this event if a connection was in the
     * middle of being established and getting ready to inform
     * the PPTP protocol engine while a disconect request was issued from
     * the PPTP protocol engine.  In this case, silently discard the
     * connection request
     */

    DBG_LEAVE();
    return PPTP_SUCCESS;

  default:
    ASSERT(!assertOnFatalError);
    DBG_LEAVE();
    return PPTP_FAILURE;

  }

  /*
   * The endpoint handle for the connection has been validated.
   * Set the control channel state to PPTP_CONTROL_CHANNEL_STARTING_SESSION.
   * Begin establishing a PPTP session to the remote peer by sending
   * a PPTP_START_SESSION_REQUEST.
   */

  controlChannelDescriptorp->controlChannelState =
    PPTP_CONTROL_CHANNEL_STARTING_SESSION;

	controlChannelDescriptorp->echoIdentNumber = 0;

  PptpBuildControlHeader(&startSessionRequest.header,
			 PPTP_START_SESSION_REQUEST,
			 sizeof(startSessionRequest));
  startSessionRequest.protocolVersion = htons(PPTP_PROTOCOL_VERSION);
  startSessionRequest.reserved1 = 0;
  startSessionRequest.reserved2 = 0;
  startSessionRequest.fepInfo.framingCapability =
    ntohl(PPTP_FRAME_CAP_ASYNC);
  startSessionRequest.fepInfo.bearerCapability =
    ntohl(PPTP_BEARER_CAP_ANALOG);
  startSessionRequest.fepInfo.maxChannels =
    htons(currentPptpConfig->netConfig.localNetConfig.fepInfo.maxChannels);
  startSessionRequest.fepInfo.firmwareRevision =
    htons(currentPptpConfig->netConfig.localNetConfig.fepInfo.firmwareRevision);
	strncpy(startSessionRequest.fepInfo.hostName,
		currentPptpConfig->netConfig.localNetConfig.fepInfo.hostName,
		MAX_HOSTNAME_LENGTH);
  strncpy(startSessionRequest.fepInfo.vendorString,
	  currentPptpConfig->netConfig.localNetConfig.fepInfo.vendorString,
	  MAX_VENDOR_STRING_LENGTH);

  if (PptpSendControlMessage(controlChannelDescriptorp,
			     (PptpControlPacket *) &startSessionRequest) ==
      PPTP_FAILURE) {
    /* TODO - disconnect control channel and go to next one */
    ASSERT(!assertOnFatalError);
    DBG_LEAVE();
    return PPTP_FAILURE;
  }

  /*
   * Schedule a new inactivity timer for the control channel
   */

  PptpRescheduleControlChannelInactivityTimer(controlChannelDescriptorp);

  DBG_LEAVE();
  return PPTP_SUCCESS;

}


PptpReturnCode PptpNetControlIncomingConnectionInd(
	PptpEndpointHandle	endpointHandle,
	PptpSockaddrIn		*remoteAddress
	)
{
  PptpAddressListIndex addressListIndex;
  PptpProtocolNumber remoteProtocolNumber;
  PptpControlChannelDescriptor *controlChannelDescriptorp;

  DBG_FUNC("PptpNetContorlIncomingConnectionInd", PPTP_PROTOCOL);

  DBG_ENTER();

  DBG_TRACE("Received new TCP connection from in_addr = %Xh sin_port = %Xh\n",
	    remoteAddress->sin_addr.s_addr, remoteAddress->sin_port);

  /*
   * If the authenticateIncomingControlChannelAddress configuration
   * option is set then validate the remote address is in the
   * remote address configuration table.  If this configuration
   * option is not set then do not disconnect on authentication failure.
   * The table is searched under all cases to lookup the value of
   * the remoteProtocolNumber.  If an address match is found in the
   * configuraiton table set the remoteProtocolNumber to the value of
   * the protocol number in the table.  Otherwise set
   * remoteProtocolNumber to the value used for the local network
   * configuration entry (defaults to PPTP_IP_GRE_PROTOCOL_NUMBER unless
   * overriden)
   */

  for (addressListIndex = 0 ;
       addressListIndex < PPTP_MAX_CONTROL_CHANNEL_SEEK_ADDRESS_ENTRIES ;
       addressListIndex++) {
    if (currentPptpConfig->netConfig.defaultRemoteAddressList.
	addressTable[addressListIndex].entryInUse) {
      if (!memcmp(&currentPptpConfig->netConfig.
		  defaultRemoteAddressList.
		  addressTable[addressListIndex].
		  remoteControlChannelAddress.sin_addr.s_addr,
		  &remoteAddress->sin_addr.s_addr,
		  sizeof(remoteAddress->sin_addr.s_addr))) {

	/*
	 * The address has been found in the network configuration table so
	 * set remoteProtocolNumber to the value found in the table
	 * and
	 */

	remoteProtocolNumber =
	  currentPptpConfig->netConfig.defaultRemoteAddressList.
	    addressTable[addressListIndex].remoteProtocolNumber;
	break;
      }
    }
  }
  if (addressListIndex == PPTP_MAX_CONTROL_CHANNEL_SEEK_ADDRESS_ENTRIES) {
    if (currentPptpConfig->authenticateIncomingControlChannelAddress) {
      PptpLogEvent("PptpNetControlIncomingConnectionInd(): could not \
validate remote address\n");
      PptpNetControlDisconnectReq(endpointHandle, 0);
      DBG_LEAVE();
      return PPTP_FAILURE;
    }
    else {
      remoteProtocolNumber =
	currentPptpConfig->netConfig.localNetConfig.localProtocolNumber;
    }
  }

  /*
   * Obtain an unused control channel descriptor and fill in the parameters
   * for it
   */

  if (PptpAssignNewControlChannelDescriptor(&controlChannelDescriptorp) !=
      PPTP_SUCCESS) {

    /*
     * An unused control channel could not be found so disconnect the
     * connection and return PPTP_FAILURE.
     */

    PptpNetControlDisconnectReq(endpointHandle, 0);
    DBG_LEAVE();
    return PPTP_FAILURE;
  }

  /*
   * An unused control channel has been found, so move it to the
   * PPTP_CONTROL_CHANNEL_STARTING_SESSION state and mark the
   * control channel as being remotely initiated.  Also fill
   * in the remote endpoint address information
   */

  controlChannelDescriptorp->controlChannelState =
    PPTP_CONTROL_CHANNEL_STARTING_SESSION;
  controlChannelDescriptorp->locallyInitiatedConnection = 0;
  controlChannelDescriptorp->controlChannelHandle = endpointHandle;
  controlChannelDescriptorp->remoteNetConfigEntry.
    remoteControlChannelAddress.sin_addr.s_addr =
      remoteAddress->sin_addr.s_addr;
  controlChannelDescriptorp->remoteNetConfigEntry.
    remoteControlChannelAddress.sin_port =
      remoteAddress->sin_port;
  controlChannelDescriptorp->remoteNetConfigEntry.
    remoteProtocolNumber =
      remoteProtocolNumber;
	  controlChannelDescriptorp->echoIdentNumber = 0;

  /*
   * Schedule a new inactivity timer for the control channel
   */

  PptpRescheduleControlChannelInactivityTimer(controlChannelDescriptorp);

  DBG_LEAVE();
  return PPTP_SUCCESS;

}


/*
 * PptpNetControlDisconnectInd()
 *
 * This routine is called to indicate that an existing control connection
 * has disconnected.  This can occur during any protocol state of a control
 * connection.
 */

PptpReturnCode PptpNetControlDisconnectInd(
	PptpEndpointHandle		endpointHandle
	)
{
  PptpControlChannelDescriptor *controlChannelDescriptorp;

  DBG_FUNC("PptpNetControlDisconnectInd", PPTP_PROTOCOL);

  DBG_ENTER();

  /*
   * Lookup the control channel number for the endpointHandle
   */

  if (PptpLookupControlChannelDescriptorByHandle(endpointHandle,
						 &controlChannelDescriptorp) !=
      PPTP_SUCCESS) {
	//
	// This means that the control channel has already been cleaned up
	// and
	//
//    ASSERT(!assertOnFatalError);
    DBG_LEAVE();
    return PPTP_FAILURE;
  }

  /*
   * Make sure the control channel is in a valid state for a disconnect
   * indication.
   */

  switch (controlChannelDescriptorp->controlChannelState) {

  case PPTP_CONTROL_CHANNEL_CONNECTING:
  case PPTP_CONTROL_CHANNEL_STARTING_SESSION:
  case PPTP_CONTROL_CHANNEL_ESTABLISHED:
  case PPTP_CONTROL_CHANNEL_STOPPING_SESSION:
  case PPTP_CONTROL_CHANNEL_DISCONNECTING:
    break;

  default:
//    ASSERT(!assertOnFatalError);
    break;

  }

  /*
   * Move the control channel to the PPTP_CONTROL_CHANNEL_LOST
   * state until all the devices using the control channel have been shut down
   */

  controlChannelDescriptorp->controlChannelState =
    PPTP_CONTROL_CHANNEL_LOST;

  /*
   * Inform all the call lines using this control channel that the
   * control channel has been remotely disconnected
   */

  PptpControlChannelRemotelyDisconnected(controlChannelDescriptorp);

  /*
   * Release the control channel descriptor since it is no longer being used
   */

  ASSERT(!controlChannelDescriptorp->inactivityTimer.timerScheduled);
  controlChannelDescriptorp->controlChannelState =
    PPTP_CONTROL_CHANNEL_NO_STATE;

  DBG_LEAVE();
  return PPTP_SUCCESS;

}


/*
 * PptpNetControlRecvDataInd()
 *
 * This routine is upon reception of incoming data from a remote endpoint
 * within the control channel.
 */

VOID PptpNetControlRecvDataInd(
	PptpEndpointHandle	controlHandle,
	UCHAR			*data,
	ULONG			dataLength
	)
{
  PptpControlChannelDescriptor	*controlChannelDescriptorp;
  PptpControlTCPDescriptor	*TCPDescriptorp;
  ULONG		 		copyLength;

  DBG_FUNC("PptpNetControlRecvDataInd", PPTP_PROTOCOL);

  DBG_ENTER();

  if (PptpLookupControlChannelDescriptorByHandle(controlHandle,
						 &controlChannelDescriptorp) !=
      PPTP_SUCCESS) {
    ASSERT(!assertOnFatalError);
    DBG_LEAVE();
    return;
  }
  TCPDescriptorp = &controlChannelDescriptorp->controlTCPDescriptor;

  while (dataLength > 0) {

    /*
     * First check to see if the current packet has its header currently
     * filled in and is awaiting the remaining data portion of the packet.
     * If not then fill in just the header portion so the message length
     * can be determined before copying in the remaining data.
     */

    if (TCPDescriptorp->currentPacketLength <
	sizeof(PptpControlMessageHeader)) {

      /*
       * The header has not been completely filled in.
       * Attempt to completely fill in the header if there
       * is enough data.
       */

      copyLength = min(dataLength,
		       sizeof(PptpControlMessageHeader) -
		       TCPDescriptorp->currentPacketLength);
      memcpy((char *) &TCPDescriptorp->currentPacket +
	     TCPDescriptorp->currentPacketLength,
	     data, copyLength);
      TCPDescriptorp->currentPacketLength += copyLength;
      data += copyLength;
      dataLength -= copyLength;

      /*
       * Check to see if the header has now been completely filled.
       * If so, then convert the header fields to host byte order.
       */

      if (TCPDescriptorp->currentPacketLength >=
	  sizeof(PptpControlMessageHeader)) {
	ASSERT(ntohl(TCPDescriptorp->currentPacket.header.magicCookie) ==
	       PPTP_MAGIC_COOKIE);
      }
      else {
	ASSERT(dataLength == 0);
      }

    }

    /*
     * If the current message header has been completely filled in
     * then the message length can be determined.  Copy in any
     * remaining data until the message has been completed as determined
     * by the length.  Once the message has been completely assembeled,
     * call PptpProtocolControlInput() to process the message.
     */

    if (TCPDescriptorp->currentPacketLength >=
	sizeof(PptpControlMessageHeader)) {
      ASSERT(ntohs(TCPDescriptorp->currentPacket.header.packetLength) >=
	     sizeof(PptpControlMessageHeader));
      ASSERT(ntohs(TCPDescriptorp->currentPacket.header.packetLength) <=
	     sizeof(PptpControlPacket));
      copyLength =
	min(dataLength,
	    ntohs(TCPDescriptorp->currentPacket.header.packetLength) -
	    TCPDescriptorp->currentPacketLength);
      memcpy((char *) &TCPDescriptorp->currentPacket +
	     TCPDescriptorp->currentPacketLength,
	     data, copyLength);
      TCPDescriptorp->currentPacketLength += copyLength;
      ASSERT(TCPDescriptorp->currentPacketLength <=
	     (ULONG) ntohs(TCPDescriptorp->currentPacket.header.packetLength));
      data += copyLength;
      dataLength -= copyLength;

      if (TCPDescriptorp->currentPacketLength ==
	  (ULONG) ntohs(TCPDescriptorp->currentPacket.header.packetLength)) {

	/*
	 * A complete control message has been assembled so call
	 * PptpProtocolControlInput() to process the message.
	 */

	PptpProtocolControlInput(controlChannelDescriptorp,
				 &TCPDescriptorp->currentPacket);

	/*
	 * Zero out the assembled control message information since the
	 * control message has been processed.
	 */

	TCPDescriptorp->currentPacketLength = 0;

      }
    }
    else {
      ASSERT(dataLength == 0);

      /*
       * We did not get a whole TCP packet, for now let's ASSERT
       * since it is probably a problem
       */

      ASSERT(!assertOnFatalError);
    }

  }

  ASSERT(dataLength == 0);
  DBG_LEAVE();

}


/*
 * PptpNetDataChannelRecvDataInd()
 *
 * This routine is called upon reception of an incoming device packet
 * from the remote endpoint.
 */

PptpReturnCode PptpNetDataChannelRecvDataInd(
	PptpSockaddrIn		*remoteAddr,
	PptpProtocolNumber	remoteProtocolNumber,
	PptpDataDescriptor	*dataDescriptorp
	)
{
	PptpCallLineDescriptor *callLineDescriptorp;
	PptpSeqNumber nextRecvDataPacketTailIndex;
	BOOLEAN	CompleteRecv = FALSE;

	DBG_FUNC("PptpNetDataChannelRecvDataInd", PPTP_PROTOCOL);

	DBG_ENTER();

	/*
	 * Call PptpTraceDataPacket to log the packet
	 */

	PptpTraceDataPacket(dataDescriptorp->callLineNumber,
				dataDescriptorp->seqNumber,
				dataDescriptorp->ackNumber,
				dataDescriptorp->data,
				dataDescriptorp->dataLength,
				1);

	/*
	 * Lookup the call line information for the device represented by
	 * callLineNumber.  The call may have been disconnected before this
	 * data packet was received so just discard it if this is the case
	 */

	if (PptpCallLineNumberToCallLineDescriptorp(dataDescriptorp->callLineNumber,
							&callLineDescriptorp) != PPTP_SUCCESS) {
#if DBG
		DbgPrint("RASPPTPE: Recv dumped, Invalid CallLineNumber %d\n",
        dataDescriptorp->callLineNumber);
#endif
		DBG_LEAVE();
		return PPTP_FAILURE;
	}

	/*
	 * Lock this call line until the packet is delivered
	 * Verify the call is in the PPTP_CALL_LINE_CONNECTED state
	 * The call may have been disconnected before this data packet was
	 * received.  Just discard it if this is the case.
	 */

	PPTP_LOCK_CALL_LINE(callLineDescriptorp->lock);
	if (callLineDescriptorp->callLineState != PPTP_CALL_LINE_CONNECTED) {
	  DBG_LEAVE();
	  PPTP_UNLOCK_CALL_LINE(callLineDescriptorp->lock);
#if DBG
	  DbgPrint("RASPPTPE: Recv dumped, Invalid CallLineState %d\n",
      callLineDescriptorp->callLineState);
#endif
	  return PPTP_FAILURE;
	}

	/*
	 * Verify that the control channel for this call is currently established.
	 * The packet may have come for a recently disconnected session
	 * or the packet may be invalid so discard it if this is the case
	 */

	PPTP_LOCK_CONTROL_CHANNEL(callLineDescriptorp->
				  controlChannelDescriptorp->lock);
	if (callLineDescriptorp->controlChannelDescriptorp->controlChannelState !=
		PPTP_CONTROL_CHANNEL_ESTABLISHED) {
	  DBG_LEAVE();
	  PPTP_UNLOCK_CONTROL_CHANNEL(callLineDescriptorp->
				  controlChannelDescriptorp->lock);
	  PPTP_UNLOCK_CALL_LINE(callLineDescriptorp->lock);
#if DBG
	  DbgPrint("RASPPTPE: Recv dumped, Invalid ControlChannelState %d\n",
      callLineDescriptorp->controlChannelDescriptorp->controlChannelState);
#endif
	  return PPTP_FAILURE;
	}

	/*
	 * Verify that the remote address of the control channel for this line
	 * matches the remote address that sent this packet
	 */

	if (memcmp(&callLineDescriptorp->controlChannelDescriptorp->
		   remoteNetConfigEntry.
		   remoteControlChannelAddress.sin_addr.s_addr,
		   &remoteAddr->sin_addr.s_addr,
		   sizeof(remoteAddr->sin_addr.s_addr)) ||
		callLineDescriptorp->controlChannelDescriptorp->
		remoteNetConfigEntry.remoteProtocolNumber !=
		remoteProtocolNumber) {

	  /*
	   * The remote address that sent the data packet does not match
	   * the remote address of the control channel sending the packet so
	   * throw the packet away
	   */

//	  ASSERT(!assertOnProtocolError);
	  DBG_LEAVE();
	  PPTP_UNLOCK_CONTROL_CHANNEL(callLineDescriptorp->
				  controlChannelDescriptorp->lock);
	  PPTP_UNLOCK_CALL_LINE(callLineDescriptorp->lock);
	  return PPTP_FAILURE;
	}
	PPTP_UNLOCK_CONTROL_CHANNEL(controlChannelDescriptorp->lock);

	/*
	 * If there is any data in the received packet then check to see
	 * if the packet window is full and if so throw away the packet
	 * If there is space left in the packet window then add the packet
	 * to the receive data packet list.
	 */

	if (dataDescriptorp->seqNumberPresent &&
		dataDescriptorp->dataLength > 0) {

		/*
		 * Check to see if the recieved sequence number is less than
		 * the highest received sequence number.  If so then we have
		 * received the packet out of order so do not pass it onto the
		 * PPP protocol since the PPTP spec stipulates that PPTP will
		 * not reorder packets.
		 */

		if ((LONG)(dataDescriptorp->seqNumber -
			callLineDescriptorp->highestInputSeqNumber) > 0) {

			callLineDescriptorp->highestInputSeqNumber =
            dataDescriptorp->seqNumber;

			nextRecvDataPacketTailIndex = callLineDescriptorp->recvDataPacketTailIndex;

			//
			// Is queue full?
			//
			if ((nextRecvDataPacketTailIndex ==
				callLineDescriptorp->recvDataPacketHeadIndex) &&
				(callLineDescriptorp->recvDataPacketArray
				[nextRecvDataPacketTailIndex].completionArg != NULL) ) {
		
				/*
				 * There is no room in the receive packet window for the packet
				 * since the window is full so throw the packet away
				 */
#if DBG
				DbgPrint("RASPPTPE: recv packet window full data dumped\n");
#endif
				callLineDescriptorp->numberDiscardedRecvPackets++;
				DBG_LEAVE();
				PPTP_UNLOCK_CALL_LINE(callLineDescriptorp->lock);
				return PPTP_FAILURE;
			}
	
			callLineDescriptorp->recvDataPacketArray
			[nextRecvDataPacketTailIndex].seqNumber = dataDescriptorp->seqNumber;
			callLineDescriptorp->recvDataPacketArray
			[nextRecvDataPacketTailIndex].data = dataDescriptorp->data;
			callLineDescriptorp->recvDataPacketArray
			[nextRecvDataPacketTailIndex].dataLength = dataDescriptorp->dataLength;
			callLineDescriptorp->recvDataPacketArray
			[nextRecvDataPacketTailIndex].completionArg = dataDescriptorp->completionArg;
		
			/*
			 * Update the receive data packet tail index
			 */
		
			if (++nextRecvDataPacketTailIndex == PPTP_MAX_PACKET_WINDOW) {
				nextRecvDataPacketTailIndex = 0;
			}
			callLineDescriptorp->recvDataPacketTailIndex = nextRecvDataPacketTailIndex;
		
			/*
			 * Call PptpServicePendingPacketList() to service the next pending
			 * packet
			 */
		
			PptpServicePendingPacketList(callLineDescriptorp);

		} else {
			/*
			 * An out of order packet has been detected so send
			 * an acknowledgement to the remote endpoint since it
			 * may have lost one of our acknowledgements and is retransmitting
			 */
			
			PptpAckRemoteDataChannel(callLineDescriptorp);

			//
			// Since this receive is being dropped we need to complete
			// the receive now
			//
			CompleteRecv = TRUE;
#if DBG
			DbgPrint("RASPPTPE: rooo cl: # %d r %8.8x hi %8.8x hp %8.8x\n",
            callLineDescriptorp->localCallLineNumber,
            dataDescriptorp->seqNumber,
			callLineDescriptorp->highestInputSeqNumber,
			callLineDescriptorp->highestProcessedInputSeqNumber);
#endif
		}
	
	} /* seqNumberPresent && datalength == 0 */

	if (dataDescriptorp->ackNumberPresent) {

		/*
		 * If the acknowledgement number is sane and greater than the previous
		 * max received acknowledgement number then update the max received
		 * acknowledgement number.  We don't check the sanity of sequence numbers
		 * before this since some implementations of the protocol may not
		 * correctly support flow control (i.e. flow control is disabled)
		 * Verify that the acknowledgement number is sane
		 */
		 if ((LONG)(dataDescriptorp->ackNumber -
			 callLineDescriptorp->highestAckedOutputSeqNumber) > 0) {

             //
             // Update the remote's receive window
             //
             callLineDescriptorp->remoteRxPacketWindow +=
             (USHORT)(dataDescriptorp->ackNumber - callLineDescriptorp->highestAckedOutputSeqNumber);
             if (callLineDescriptorp->remoteRxPacketWindow >
                 callLineDescriptorp->remoteRxPacketWindowMax) {
                 callLineDescriptorp->remoteRxPacketWindow =
                 callLineDescriptorp->remoteRxPacketWindowMax;
             }

			 callLineDescriptorp->highestAckedOutputSeqNumber = dataDescriptorp->ackNumber;
		 }

		ASSERT(callLineDescriptorp->lastOutputSeqNumber >=
		callLineDescriptorp->highestAckedOutputSeqNumber);

		/*
		 * If the acknowledgment number indicates that there is space for
		 * more packets in the window and the data channel was previously
		 * flow controlled to the remote endpoint then resume sending packets
		 * through the data channel
		 */

		if (callLineDescriptorp->remoteDataChannelFlowControlOn &&
			((LONG)(callLineDescriptorp->lastOutputSeqNumber -
			callLineDescriptorp->highestAckedOutputSeqNumber) <
			callLineDescriptorp->remoteRxPacketWindow)) {
			PptpReturnCode status;

			callLineDescriptorp->remoteDataChannelFlowControlOn = 0;

			switch (currentPptpConfig->dataChannelFlowControlType) {
				case PPTP_DATA_CHANNEL_FLOW_CONTROL_STATIC_TIMEOUT:
				//
				// We must have a timer scheduled so cancel it
				//
				PptpUnscheduleInternalTimer(&callLineDescriptorp->flowControlTimer);
				break;

				default:
				break;
			}
#if defined(PPTP_SERVER)
			status = PptpServerDataChannelSendAvailable(callLineDescriptorp);
#else
			status = PptpClientDataChannelSendAvailable(callLineDescriptorp);
#endif /* defined(PPTP_SERVER) */
			if (status != PPTP_SUCCESS) {
				ASSERT(!assertOnFatalError);
			}
		}
	}

	/*
	 * If datalength is 0 then the packet is an ack that carries no
	 * data with it, so call PptpNetDataChannelRecvDataCompleteReq() to
	 * inform the network library that the received buffer is no longer
	 * required.  If this packet is being dropped we will need to complete
	 * the receive now.
	 */
	if (dataDescriptorp->dataLength == 0 ||
		CompleteRecv) {

		if (PptpNetDataChannelRecvDataCompleteReq(dataDescriptorp->completionArg) != PPTP_SUCCESS) {
			ASSERT(!assertOnFatalError);
		}

	} // datalength == 0

	DBG_LEAVE();
	PPTP_UNLOCK_CALL_LINE(callLineDescriptop->lock);
	return PPTP_SUCCESS;
}


/*
 * PptpNetDataChannelSendDataCompleteInd()
 *
 * This routine is called to notify the PPTP engine that the handling of
 * an outbound packet for a remote endpoint has been completed and its
 * associated data buffer is no longer required.
 */

PptpReturnCode PptpNetDataChannelSendDataCompleteInd(
	VOID 				*completionArg
	)
{
  PptpReturnCode status;

  DBG_FUNC("PptpNetDataChannelSendDataCompleteInd", PPTP_PROTOCOL);

  DBG_ENTER();

#if defined(PPTP_SERVER)
  status = PptpServerDataChannelSendComplete(completionArg);
#else
  status = PptpClientDataChannelSendComplete(completionArg);
#endif /* defined(PPTP_SERVER) */

  DBG_LEAVE();
  return status;

}


PptpReturnCode PptpAllocateControlChannelForCall(
	PptpCallLineDescriptor	*callLineDescriptorp,
	PptpAddressList		*pAddressList
	)
{
  PptpControlChannelSeekIndex controlChannelSeekIndex;
  PptpReturnCode status;

  DBG_FUNC("PptpAllocateControlChannelForCall", PPTP_PROTOCOL);

  DBG_ENTER();

  ASSERT(callLineDescriptorp->callLineState == PPTP_CALL_LINE_ALLOCATED);

  /*
   * Put the call line into the PPTP_CALL_LINE_WAITING_CONTROL_CHANNEL state
   */

  callLineDescriptorp->callLineState = PPTP_CALL_LINE_WAITING_CONTROL_CHANNEL;

  /*
   * There are two types of control channel allocations to handle:
   *
   *   - A control channel allocation using the default control channel
   *     seek list.  This case is handled seperately since it contains
   *     logic that remembers which remote address in the control channel
   *     last accpeted a call and reuses this address for new calls until
   *     that endpoint no longer accepts calls.  At that point it tries
   *     to use any existing control channel to establish the call and
   *     if that fails it then tries to establish control channels with
   *     any host in the default control channel list.  Note that the
   *     zeroth entry in the control channel seek descriptor table is
   *     reserved for the default control channel seek descriptor
   *
   *   - The case where a specified control channel seek list is used
   *     for just this call only.  For this scenario a host will first
   *     use any existing control channels to establish the call and
   *     if that fails it then tries to establish a control channel
   *     with any host in the default control channel list.
   */

  if (pAddressList == NULL) {

    /*
     * The call is using the default control channel so point the
     * call control channel seek descriptor to the seek descriptor
     * default control channel seek descriptor
     */

    controlChannelSeekIndex = 0;
    callLineDescriptorp->controlChannelSeekDescriptorp =
      &controlChannelSeekDescriptorTable[0];

    /*
     * If a control channel seek for the default control channel
     * is currently in progress then return and wait to be woken
     * up later.  Otherwise, initialize the seek indices for the default
     * control channel seek descriptor to point to the address list
     * index for the last established address list index in the
     * seek descriptor
     */

    if (controlChannelSeekDescriptorTable[0].controlChannelSeekState !=
	PPTP_SEEK_NO_STATE) {
      DBG_LEAVE();
      return PPTP_SUCCESS;
    }
    else {
      controlChannelSeekDescriptorTable[0].currentAddressListIndex =
	controlChannelSeekDescriptorTable[0].finalAddressListIndex =
	  controlChannelSeekDescriptorTable[0].lastEstablishedAddressListIndex;
    }

  }
  else {

    /*
     * The call is using a specified control channel seek list so first
     * search for an unused control channel seek descriptor to use until
     * the call is established
     */

    for (controlChannelSeekIndex = 1 ;
	 controlChannelSeekIndex <
	 PPTP_MAX_CONTROL_CHANNEL_SEEK_NUMBER ;
	 controlChannelSeekIndex++) {
      if (controlChannelSeekDescriptorTable[controlChannelSeekIndex].
	  controlChannelSeekState ==
	  PPTP_SEEK_NO_STATE) {
	callLineDescriptorp->controlChannelSeekDescriptorp =
	  &controlChannelSeekDescriptorTable[controlChannelSeekIndex];
	break;
      }
    }
    if (controlChannelSeekIndex ==
	PPTP_MAX_CONTROL_CHANNEL_SEEK_NUMBER) {

      /*
       * An unused control channel seek descriptor is not available so
       * return PPTP_FAILURE
       */

      DBG_LEAVE();
      return PPTP_FAILURE;
    }

    /*
     * An unused control channel seek descriptor has been found,
     * initialize the control channel seek descriptor seek indices
     * and address list entries.  Note that the last established control
     * channel seek index is unused for a specificed control channel
     * seek list
     */

    callLineDescriptorp->controlChannelSeekDescriptorp->
      controlChannelSeekIndex = controlChannelSeekIndex;

    callLineDescriptorp->controlChannelSeekDescriptorp->
      currentAddressListIndex =
      callLineDescriptorp->controlChannelSeekDescriptorp->
	finalAddressListIndex =
	  0;

    memcpy(&callLineDescriptorp->controlChannelSeekDescriptorp->addressList,
	   pAddressList,
	   sizeof(*pAddressList));
  }

  DBG_TRACE("Start control channel seek for call = %d currentAddressIndex = %d finalAddressIndex = %d\n",
	     callLineDescriptorp->localCallLineNumber,
	     callLineDescriptorp->controlChannelSeekDescriptorp->
	     currentAddressListIndex,
	     callLineDescriptorp->controlChannelSeekDescriptorp->
	     finalAddressListIndex
	     );

  ASSERT(callLineDescriptorp->controlChannelSeekDescriptorp->
	 controlChannelSeekState == PPTP_SEEK_NO_STATE);

  /*
   * Start the control channel seek descriptor search.  If this
   * fails then a seek descriptor for the control channel could
   * not be allocated or started
   */

  status = PptpContinueControlChannelSeekDescriptorSearch(
		callLineDescriptorp->controlChannelSeekDescriptorp);
  if (status != PPTP_SUCCESS) {

    /*
     * The control channel seek descriptor could not be started so
     * unbind the call line desciptor form it and deallocate it
     */

    callLineDescriptorp->controlChannelSeekDescriptorp = NULL;
    controlChannelSeekDescriptorTable[controlChannelSeekIndex].
      controlChannelSeekState = PPTP_SEEK_NO_STATE;
  }

  DBG_LEAVE();
  return status;

}


/*
 * PptpControlChannelAcceptedCall()
 *
 * This routine is called when a control channel accepts a call
 */

PptpReturnCode PptpControlChannelAcceptedCall(
	PptpCallLineDescriptor	*callLineDescriptorp
	)
{
  PptpControlChannelSeekDescriptor *controlChannelSeekDescriptorp;

  DBG_FUNC("PptpControlChannelAcceptedCall", PPTP_PROTOCOL);

  DBG_ENTER();

  /*
   * Make sure that there is a seek descriptor for this call line
   */

  if (!callLineDescriptorp->controlChannelSeekDescriptorp) {
    ASSERT(!assertOnFatalError);
    DBG_LEAVE();
    return PPTP_FAILURE;
  }

  /*
   * Verify that the call line descriptor is now bound to the
   * seek descriptor control channel descriptor
   */

  if (callLineDescriptorp->controlChannelSeekDescriptorp->
      pendingControlChannelDescriptorp !=
      callLineDescriptorp->controlChannelDescriptorp) {
    ASSERT(!assertOnFatalError);
    DBG_LEAVE();
    return PPTP_FAILURE;
  }

  /*
   * Detach the call line from the seek descriptor
   */

  controlChannelSeekDescriptorp =
    callLineDescriptorp->controlChannelSeekDescriptorp;
  callLineDescriptorp->controlChannelSeekDescriptorp = NULL;

  /*
   * Wakeup any seek descriptors that were blocked on the control
   * channel descriptor
   */

  PptpContinuePendingCallLineDescriptors(callLineDescriptorp->
					 controlChannelDescriptorp);

  /*
   * Check to see if the seek descriptor for the call is required
   * anymore
   */

  PptpCheckControlChannelSeekDescriptorUsage(controlChannelSeekDescriptorp);

  DBG_LEAVE();
  return PPTP_SUCCESS;

}


/*
 * PptpControlChannelRefusedCall
 *
 * This routine is called whan a control channel refuses to accept
 * a requested call
 */

PptpReturnCode PptpControlChannelRefusedCall(
	PptpCallLineDescriptor	*callLineDescriptorp
	)
{
  PptpControlChannelDescriptor *previousControlChannelDescriptorp;

  DBG_FUNC("PptpControlChannelRefusedCall", PPTP_PROTOCOL);

  DBG_ENTER();

  /*
   * Verify that the call line desriptor was bound to a control
   * channel descriptor in the proper state
   */

  ASSERT(callLineDescriptorp->controlChannelDescriptorp);
  if (callLineDescriptorp->controlChannelDescriptorp->controlChannelState !=
      PPTP_CONTROL_CHANNEL_ESTABLISHED) {
    ASSERT(!assertOnFatalError);
    DBG_LEAVE();
    return PPTP_FAILURE;
  }

  /*
   * Unbind the call line descriptor from its control channel descriptor
   */

  previousControlChannelDescriptorp =
    callLineDescriptorp->controlChannelDescriptorp;
  callLineDescriptorp->controlChannelDescriptorp = NULL;

  /*
   * Wakeup any control channel seek descriptors that were waiting
   * on this control channel and have them search for a new control
   * channel to use
   */

  PptpContinuePendingControlChannelSeekDescriptors(
	previousControlChannelDescriptorp);

  /*
   * Check to see if the control channel that refused the call is
   * required anymore
   */

  PptpCheckControlChannelDescriptorUsage(previousControlChannelDescriptorp);

  DBG_LEAVE();
  return PPTP_SUCCESS;

}


/*
 * PptpControlChannelDisconnectedCall()
 *
 * This routine is called whan a control channel completes orderly
 * disconnection of a call
 */

PptpReturnCode PptpControlChannelDisconnectedCall(
	PptpCallLineDescriptor	*callLineDescriptorp
	)
{
  PptpControlChannelDescriptor *previousControlChannelDescriptorp;

  DBG_FUNC("PptpControlChannelDisconnectedCall", PPTP_PROTOCOL);

  DBG_ENTER();

  /*
   * If there was a flow control timer scheduled for the call
   * then unschedule it
   */

  if (callLineDescriptorp->flowControlTimer.timerScheduled) {
    PptpUnscheduleInternalTimer(&callLineDescriptorp->flowControlTimer);
  }

  if (callLineDescriptorp->delayedAckTimer.timerScheduled) {
	  PptpUnscheduleInternalTimer(&callLineDescriptorp->delayedAckTimer);
  }

  /*
   * Verify that the call line desriptor was bound to a control
   * channel descriptor in the proper state
   */

  ASSERT(callLineDescriptorp->controlChannelDescriptorp);
  if (callLineDescriptorp->controlChannelDescriptorp->controlChannelState !=
      PPTP_CONTROL_CHANNEL_ESTABLISHED) {
    ASSERT(!assertOnFatalError);
    DBG_LEAVE();
    return PPTP_FAILURE;
  }

  /*
   * If the call channel is bound to a seek descriptor then it
   * was disconnected locally while a call connection request
   * was pending.  For this case, we wait until the call completes
   * disconnection before allowing the control channel seek to
   * continue so here is the time to continue the control channel seek
   */

  if (callLineDescriptorp->controlChannelSeekDescriptorp) {

    /*
     * The coninuation of the control channel seek descriptor
     * here is the same as that when a call is accepted and the
     * control channel seek is continued
     */

    PptpControlChannelAcceptedCall(callLineDescriptorp);
  }

  /*
   * Unbind the call line descriptor from its control channel descriptor
   */

  previousControlChannelDescriptorp =
    callLineDescriptorp->controlChannelDescriptorp;
  callLineDescriptorp->controlChannelDescriptorp = NULL;

  /*
   * Check to see if the control channel that disconnected the call is
   * required anymore
   */

  PptpCheckControlChannelDescriptorUsage(previousControlChannelDescriptorp);

  DBG_LEAVE();
  return PPTP_SUCCESS;

}


/*
 * PptpLocallyDisconnectedCall()
 *
 * This routine is called when a call is locally disconnected while
 * it is waiting for a control channel connection
 */

PptpReturnCode PptpLocallyDisconnectedCall(
	PptpCallLineDescriptor	*callLineDescriptorp
	)
{
  PptpControlChannelSeekDescriptor *controlChannelSeekDescriptorp;

  DBG_FUNC("PptpLocallyDisconnectedCall", PPTP_PROTOCOL);

  DBG_ENTER();

  /*
   * Verify that the call is using a control channel seek desriptor
   */

  if (!callLineDescriptorp->controlChannelSeekDescriptorp) {
    ASSERT(!assertOnFatalError);
    DBG_LEAVE();
    return PPTP_FAILURE;
  }

  /*
   * Verify that the control channel seek descriptor is currently bound to a
   * control channel desriptor
   */

  if (!callLineDescriptorp->controlChannelSeekDescriptorp->
      pendingControlChannelDescriptorp) {
    ASSERT(!assertOnFatalError);
    DBG_LEAVE();
    return PPTP_FAILURE;
  }

  /*
   * Verify that the call is in the right state for this routine
   * to get called
   */

  switch (callLineDescriptorp->callLineState) {

  case PPTP_CALL_LINE_WAITING_CONTROL_CHANNEL:
  case PPTP_CALL_LINE_CALLING:
  case PPTP_CALL_LINE_OFFERING:
    break;

  default:
    ASSERT(!assertOnFatalError);
    DBG_LEAVE();
    return PPTP_FAILURE;
  }

  /*
   * Unbind the call line from its control channel seek descriptor
   */

  controlChannelSeekDescriptorp =
    callLineDescriptorp->controlChannelSeekDescriptorp;
  callLineDescriptorp->controlChannelSeekDescriptorp = NULL;

  /*
   * Check to see if the seek descriptor for the call is required
   * anymore
   */

  PptpCheckControlChannelSeekDescriptorUsage(controlChannelSeekDescriptorp);

  DBG_LEAVE();
  return PPTP_SUCCESS;

}


/*
 * PptpRelinquishCallLine()
 *
 * This routine is called to release a call line that is no longer being
 * used so it can be allocated for a new call
 */

PptpReturnCode PptpRelinquishCallLine(
	PptpCallLineDescriptor *callLineDescriptorp
)
{
  DBG_FUNC("PptpRelinquishCallLine", PPTP_PROTOCOL);

  DBG_ENTER();

  DBG_TRACE("Relinquishing call line = %d\n",
	    callLineDescriptorp->localCallLineNumber);

  /*
   * If there was a flow control timer scheduled for the call
   * then unschedule it
   */

  if (callLineDescriptorp->flowControlTimer.timerScheduled) {
    PptpUnscheduleInternalTimer(&callLineDescriptorp->flowControlTimer);
  }

  /*
   * If there was a delayed ack timer scheduled for the call
   * then unschedule it
   */

  if (callLineDescriptorp->delayedAckTimer.timerScheduled) {
    PptpUnscheduleInternalTimer(&callLineDescriptorp->
				delayedAckTimer);
  }

  /*
   * Make sure the call line is in a state that it can be relinquished from
   */

  switch (callLineDescriptorp->callLineState) {

  case PPTP_CALL_LINE_ALLOCATED:
  case PPTP_CALL_LINE_WAITING_CONTROL_CHANNEL:
  case PPTP_CALL_LINE_DISCONNECTING_LOCAL:
  case PPTP_CALL_LINE_DISCONNECTING_REMOTE:
  case PPTP_CALL_LINE_LOST_CONTROL_CHANNEL:
    break;

  default:
    ASSERT(!assertOnFatalError);
    DBG_LEAVE();
    return PPTP_FAILURE;

  }

  /*
   * Verify that the call line is not currently bound to a seek
   * descriptor
   */

  if (callLineDescriptorp->controlChannelSeekDescriptorp) {
    ASSERT(!assertOnFatalError);
    DBG_LEAVE();
    return PPTP_FAILURE;
  }

  /*
   * Verify that the call line is not currently bound to a control
   * channel descriptor
   */

  if (callLineDescriptorp->controlChannelDescriptorp) {
    ASSERT(!assertOnFatalError);
    DBG_LEAVE();
    return PPTP_FAILURE;
  }

  /*
   * Set the call line state to PPTP_CALL_LINE_NO_STATE
   */

  callLineDescriptorp->callLineState = PPTP_CALL_LINE_NO_STATE;

  DBG_LEAVE();
  return PPTP_SUCCESS;

}


/*
 * PptpContinueControlChannelSeekDescriptorSearch()
 *
 * This routine forces a seek descriptor search for a control channel
 * to continue with the next control channel.  This routine is called
 * when one of the following happens:
 *
 *  - a control channel search has just started
 *
 *  - a control channel being used by a seek descriptor was
 *    disconnected
 *
 *  - a control channel refused to accept a call
 *
 * If the control channel search for the seek descriptor cannot continue
 * then the seek is aborted for this seek descriptor.
 */

STATIC PptpReturnCode PptpContinueControlChannelSeekDescriptorSearch(
	PptpControlChannelSeekDescriptor *controlChannelSeekDescriptorp
	)
{
  PptpControlChannelNumber controlChannelNumber;

  DBG_FUNC("PptpContinueControlChannelSeek", PPTP_PROTOCOL);

  DBG_ENTER();

  DBG_TRACE("Continue control channel seek, seek state = %d, currentAddressIndex = %d finalAddressIndex = %d\n",
	    controlChannelSeekDescriptorp->controlChannelSeekState,
	    controlChannelSeekDescriptorp->
	    currentAddressListIndex,
	    controlChannelSeekDescriptorp->
	    finalAddressListIndex
	    );
	
  /*
   * There are two phases used when seeking for a control channel
   * to use to establish a call with:
   *
   *   - First cycle through the control channel address list, searching
   *     for an entry that matches the address of any established PPTP
   *     sessions in the control channel table.  Attempt to establish calls
   *     through these established control channels.
   *
   *   - If the first phase failed, then next cycle through the control
   *     channel address list again, this time seraching for entries that
   *     do not match the address of any established PPTP sessions in
   *     the control channel table.  Attempt to first establish a control
   *     channel and then a call with these addresses.
   */

  switch (controlChannelSeekDescriptorp->controlChannelSeekState) {

  case PPTP_SEEK_NO_STATE:
  case PPTP_SEEKING_OUTSTANDING_CHANNEL:

    /*
     * This is the first phase of control channel seeking
     */

    while (1) {

      if (controlChannelSeekDescriptorp->controlChannelSeekState ==
	  PPTP_SEEK_NO_STATE) {

	/*
	 * This is the first iteration of the seek cycle so update
	 * the seek state so the next time this address list index
	 * is reached we know the cycle has completed and will
	 * jump to the seond phase of control channel seeking
	 */

	controlChannelSeekDescriptorp->controlChannelSeekState =
	  PPTP_SEEKING_OUTSTANDING_CHANNEL;
      }
      else {

	/*
	 * This is a continuation of a seek cycle during this phase
	 * of control channel seeking, so check to see if there are
	 * any remaining entries that have not been tried in the
	 * control channel address list that we currently have
	 * control channels established with.  If the seek has ended
	 * then jump to the next phase, otherwise continue the seeking
	 * within this phase.  First cycle to the next control
	 * channel address in the control channel seek descriptor,
	 * remembering to check for wrap-around
	 */

	controlChannelSeekDescriptorp->currentAddressListIndex++;
	if (controlChannelSeekDescriptorp->currentAddressListIndex ==
	    PPTP_MAX_CONTROL_CHANNEL_SEEK_ADDRESS_ENTRIES) {
	  controlChannelSeekDescriptorp->currentAddressListIndex = 0;
	}

	if (controlChannelSeekDescriptorp->currentAddressListIndex ==
	    controlChannelSeekDescriptorp->finalAddressListIndex) {

	  /*
	   * All the entries have been cycled through for the first phase
	   * and no currently established control channel could be
	   * found that would accept the call so break out of the
	   * loop and jump to the next phase of control channel seeking
	   */

	  break;
	}

      }

      /*
       * Check to see if there is an entry in the control channel
       * descriptor table with an address that matches the current control
       * channel address seek descriptor entry for any control channel
       * that is either already established or is in the control channel
       * establishment phase.
       */

      for (controlChannelNumber = 0 ;
	   controlChannelNumber < currentPptpConfig->numberLineDevices ;
	   controlChannelNumber++) {

	if (PptpControlChannelAddressMatch(&controlChannelDescriptorTable
					   [controlChannelNumber].
					   remoteNetConfigEntry,
					   &controlChannelSeekDescriptorp->
					   addressList.
					   addressTable
					   [controlChannelSeekDescriptorp->
					    currentAddressListIndex]) ==
	    PPTP_SUCCESS) {

	  switch (controlChannelDescriptorTable[controlChannelNumber].
		  controlChannelState) {

	  case PPTP_CONTROL_CHANNEL_CONNECTING:
	  case PPTP_CONTROL_CHANNEL_STARTING_SESSION:

	    /*
	     * An address match has been found for a control channel
	     * that is currently in the phase of control channel
	     * establishment.  Wait until the reosolution of this
	     * control channel is available to determine whether or not
	     * to use it for the call
	     */

		DBG_TRACE("Seek waiting for pending control channel\n");

	    controlChannelSeekDescriptorp->pendingControlChannelDescriptorp =
	      &controlChannelDescriptorTable[controlChannelNumber];

	    DBG_LEAVE();
	    return PPTP_SUCCESS;

	  case PPTP_CONTROL_CHANNEL_ESTABLISHED: {
	    PptpControlChannelSeekIndex controlChannelSeekIndex;

	    /*
	     * An address match has been found for a control channel
	     * that is currently established so attempt to use the
	     * control channel
	     */

	    DBG_TRACE("Seek using established control channel\n");

	    controlChannelSeekDescriptorp->pendingControlChannelDescriptorp =
	      &controlChannelDescriptorTable[controlChannelNumber];

	    /*
	     * Check to see if any other seek descriptors are currently
	     * using this control channel, if so then a call is in progress
	     * on the control channel.  If no call is in progress wihtin
	     * the control channel then wakeup any pending call descriptors
	     * for the control channel (the only one woken up should
	     * be this new one)
	     */

		for (controlChannelSeekIndex = 0 ;
			 controlChannelSeekIndex < PPTP_MAX_CONTROL_CHANNEL_SEEK_NUMBER ;
			 controlChannelSeekIndex++) {
			if ((controlChannelSeekDescriptorp !=
				&controlChannelSeekDescriptorTable[controlChannelSeekIndex]) &&
				(controlChannelSeekDescriptorTable[controlChannelSeekIndex].
				controlChannelSeekState != PPTP_SEEK_NO_STATE) &&
				(controlChannelSeekDescriptorTable[controlChannelSeekIndex].
				pendingControlChannelDescriptorp ==
				&controlChannelDescriptorTable[controlChannelNumber])) {
				break;
			}
		}

		if (controlChannelSeekIndex == PPTP_MAX_CONTROL_CHANNEL_SEEK_NUMBER) {

	      /*
	       * There are no seek descriptors using the control channel
	       * so wakeup any pending call line descriptors for the
	       * control channel
	       */
	
			if (PptpContinuePendingCallLineDescriptors(
				controlChannelSeekDescriptorp->
				pendingControlChannelDescriptorp) != PPTP_SUCCESS) {
				ASSERT(!assertOnFatalError);
				DBG_LEAVE();
				return PPTP_FAILURE;
			}
	    }

	    DBG_LEAVE();
	    return PPTP_SUCCESS;
	  }

	  default:

	    /*
	     * An address match was found for a control channel that not
	     * in already established or in the establishment phase so
	     * cycle to the next entry in the control channel
	     * address list
	     */

	    break;

	  } /* end switch */
	} /* end if */
      } /* end for */
    } /* end while(1) */

    DBG_TRACE("Seek found no current control control\n");

    /*
     * The first phase of control channel establishment has completed
     * and there were not any established control channels that would
     * accept the call so begin the second phase of control channel
     * seeking where an attempt to establish.
     */

    ASSERT(controlChannelSeekDescriptorp->currentAddressListIndex ==
	   controlChannelSeekDescriptorp->finalAddressListIndex);

    /* fall through to the PPTP_SEEKING_NEW_CHANNEL state */

  case PPTP_SEEKING_NEW_CHANNEL: {

    /*
     * This is the seoncd phase of control channel establishment.
     * Cycle through the entries in the control channel address list
     * attempting to establish first control channels and then calls
     * with those addresses
     */

    while (1) {
      BOOL correspondingControlChannelInUse = 0;

      if (controlChannelSeekDescriptorp->controlChannelSeekState ==
	  PPTP_SEEKING_OUTSTANDING_CHANNEL) {

	/*
	 * This is the first iteration of the cycle for this phase to
	 * update the current seek state so the next time this config index
	 * is reached we know the cycle has completed
	 * Note that this code is entered only from the fall through
	 * case of the PPTP_SEEKING_OUTSTANDING_CHANNEL case
	 */

	controlChannelSeekDescriptorp->controlChannelSeekState =
	  PPTP_SEEKING_NEW_CHANNEL;

      }
      else {

	/*
	 * This is a continuation of a seek cycle during this phase
	 * of control channel seeking, so cycle to the next entry in
	 * control channel address list that is not in a phase of control
	 * channel establishment, establishing a control channel with that
	 * address and then a call with the established control channel
	 * Once the seek cycle ends, the call cannot be established with
	 * any of the entries in the control channel address list so
	 * fail the call.  First cycle to the next entry in the
	 * control channel address list, checking for wraparound
	 */
	
	controlChannelSeekDescriptorp->currentAddressListIndex++;
	if (controlChannelSeekDescriptorp->currentAddressListIndex ==
	    PPTP_MAX_CONTROL_CHANNEL_SEEK_ADDRESS_ENTRIES) {
	  controlChannelSeekDescriptorp->currentAddressListIndex = 0;
	}

	if (controlChannelSeekDescriptorp->currentAddressListIndex ==
	    controlChannelSeekDescriptorp->finalAddressListIndex) {

	  /*
	   * All the entries have been cycled through and no control
	   * channel could be found to accept the call so break out of the
	   * loop
	   */

	  break;
	}

      }

      /*
       * Check to see if the current entry in the control channel address
       * list is valid, if not skip to the next entry
       */

      if (!controlChannelSeekDescriptorp->addressList.addressTable

	  [controlChannelSeekDescriptorp->currentAddressListIndex].
	  entryInUse) {
	continue;
      }

      /*
       * Make sure that this address is not currently in a phase
       * of control channel establishement.  If it is then skip to
       * the next entry
       */

      for (controlChannelNumber = 0 ;
	   controlChannelNumber < currentPptpConfig->numberLineDevices ;
	   controlChannelNumber++) {

	switch (controlChannelDescriptorTable[controlChannelNumber].
		controlChannelState) {

	case PPTP_CONTROL_CHANNEL_NO_STATE:
	case PPTP_CONTROL_CHANNEL_ALLOCATED:
	  break;

	default:

	  if (PptpControlChannelAddressMatch(&controlChannelDescriptorTable
					     [controlChannelNumber].
					     remoteNetConfigEntry,
					     &controlChannelSeekDescriptorp->
					     addressList.
					     addressTable
					     [controlChannelSeekDescriptorp->
					      currentAddressListIndex]) ==
	      PPTP_SUCCESS) {
	    correspondingControlChannelInUse = 1;
	  }
	  break;
	}
      }
      if (correspondingControlChannelInUse) {
	continue;
      }

      /*
       * Find an unused control channel to use for the new connection
       */

      if (PptpAssignNewControlChannelDescriptor(
		&controlChannelSeekDescriptorp->
		pendingControlChannelDescriptorp) != PPTP_SUCCESS) {

	/*
	 * An unused control channel could not be found so return
	 * PPTP_FAILURE.  The caller will abort the control channel
	 * seek descriptor search if necessary.
	 *
	 * This is probably an error where control channels are getting
	 * lost so ASSERT as a fatal error here as well
	 */

	ASSERT(!assertOnFatalError);
	DBG_LEAVE();
	return PPTP_FAILURE;
      }

      /*
       * An unused control channel has been located.
       * Fill in the allocated control channel descriptor for this call
       * with the current address in the control channel address list
       */

      DBG_TRACE("Seek allocating unused control channel = %d\n",
		controlChannelSeekDescriptorp->
		pendingControlChannelDescriptorp->controlChannelIndex);

      PptpSetControlChannelAddress(&controlChannelSeekDescriptorp->
				   pendingControlChannelDescriptorp->
				   remoteNetConfigEntry,
				   &controlChannelSeekDescriptorp->
				   addressList.
				   addressTable
				   [controlChannelSeekDescriptorp->
				    currentAddressListIndex]);

      if (PptpAttemptControlChannelConnection(
		controlChannelSeekDescriptorp->
		pendingControlChannelDescriptorp) ==
	  PPTP_SUCCESS) {
	DBG_LEAVE();
	return PPTP_SUCCESS;
      }

    }

    /*
     * None of the entries in the control channel address list would
     * accept this call during this phase of control channel seeking,
     * so return PPTP_FAILURE.  The caller will abort the control channel
     * seek descriptor search if required.
     */

    DBG_TRACE("Seek failed\n");

    DBG_LEAVE();
    return PPTP_FAILURE;
  }

  default:
    ASSERT(!assertOnFatalError);
    DBG_LEAVE();
    return PPTP_FAILURE;

  }

}


/*
 * PptpAbortControlChannelSeekDescriptorSearch()
 *
 * This routine is called to abort a control channel search in progress
 * for a particular control channel seek descriptor.  It can get
 * called when one of the following happens:
 *
 *   - An unused control channel descriptor could not be located
 *
 *   - The seek has failed since it was unable to find a control
 *     channel to accept the call
 *
 *   - A call line disconnected before it was accepted and there
 *     are not more call lines using the control channel seek
 *     descriptor
 */

STATIC PptpReturnCode PptpAbortControlChannelSeekDescriptorSearch(
	PptpControlChannelSeekDescriptor *controlChannelSeekDescriptorp
	)
{
  PptpCallLineNumber callLineNumber;

  DBG_FUNC("PptpAbortControlChannelSeekDescriptorSearch", PPTP_PROTOCOL);

  DBG_ENTER();

  DBG_TRACE("Aborting control channel seek search\n");

  /*
   * Since the control channel seek descriptor is aborting, the control
   * channel seek descriptor should not be using a control channel, so
   * verify this
   */

  if (controlChannelSeekDescriptorp->pendingControlChannelDescriptorp) {
    ASSERT(!assertOnFatalError);
    DBG_LEAVE();
    return PPTP_FAILURE;
  }

  /*
   * Go through each call line that was waiting on this control
   * channel seek descriptor and inform the device that a call
   * could not be established.  This is accomplished by informing the
   * call line that it lost its control channel.
   */

  for (callLineNumber = 0 ;
       callLineNumber < currentPptpConfig->numberLineDevices;
       callLineNumber++) {

    if (callLineDescriptorTable[callLineNumber].
	controlChannelSeekDescriptorp == controlChannelSeekDescriptorp) {

      ASSERT(callLineDescriptorTable[callLineNumber].
	     controlChannelSeekDescriptorp);

      if (callLineDescriptorTable[callLineNumber].
	  controlChannelSeekDescriptorp == controlChannelSeekDescriptorp) {
	PptpReturnCode status;

	/*
	 * The call line no longer is using the control channel seek
	 * descriptor so sever its connection to it
	 */

	callLineDescriptorTable[callLineNumber].controlChannelSeekDescriptorp =
	  NULL;

#if defined(PPTP_SERVER)
	status =
	  PptpServerCallLineLostControlChannel(&callLineDescriptorTable
					       [callLineNumber]);
#else
	status =
	  PptpClientCallLineLostControlChannel(&callLineDescriptorTable
					       [callLineNumber]);
#endif /* defined(PPTP_SERVER) */

      ASSERT(status == PPTP_SUCCESS);
      }
    }
  }

  /*
   * The seek has been aborted so release the seek descriptor
   */

  controlChannelSeekDescriptorp->controlChannelSeekState = PPTP_SEEK_NO_STATE;

  DBG_LEAVE();
  return PPTP_SUCCESS;

}


/*
 * PptpContinuePendingCallLineDescriptors()
 *
 * This routine is called to cause the next pending call line descriptor
 * whose seek descriptor waiting to use this control channel to begin
 * using the call line descriptor to establish a call.  This routine
 * is called when one of the following events happens:
 *
 *  - a control channel has accepted a pending call
 *
 *  - a call line requesting use of a call line to accept a call
 *    has disconnected the call before it was accepted
 */

STATIC PptpReturnCode PptpContinuePendingCallLineDescriptors(
	PptpControlChannelDescriptor	*controlChannelDescriptorp
	)
{
  PptpCallLineNumber callLineNumber;

  DBG_FUNC("PptpContinuePendingCallLineDescriptors",
	   PPTP_PROTOCOL);

  DBG_ENTER();

  /*
   * Make sure the control channel descriptor is in the proper state
   * to start assigning calls to it
   */

  if (controlChannelDescriptorp->controlChannelState !=
      PPTP_CONTROL_CHANNEL_ESTABLISHED) {
    ASSERT(!assertOnFatalError);
    DBG_LEAVE();
    return PPTP_FAILURE;
  }

  /*
   * Search through callLineDescriptorTable searching for a control channel
   * whose control channel seek descriptor is pending on a control channel
   * descriptor that matches the one the call lines are being woken up for.
   * When a match is found, attempt to establish that call and wait
   * for a response before waking up any more pending calls
   */

  for (callLineNumber = 0 ;
       callLineNumber < currentPptpConfig->numberLineDevices ;
       callLineNumber++) {

    if (callLineDescriptorTable[callLineNumber].
	controlChannelSeekDescriptorp) {

      /*
       * Make sure the call is in a proper state to be using a seek
       * descriptor
       */

      if (callLineDescriptorTable[callLineNumber].callLineState !=
	  PPTP_CALL_LINE_WAITING_CONTROL_CHANNEL) {
	ASSERT(!assertOnFatalError);
	continue;
      }

      ASSERT(callLineDescriptorTable[callLineNumber].
	     controlChannelSeekDescriptorp->pendingControlChannelDescriptorp);

      if (callLineDescriptorTable[callLineNumber].
	  controlChannelSeekDescriptorp->pendingControlChannelDescriptorp ==
	  controlChannelDescriptorp) {

	/*
	 * Assign the call line to the control channel descriptor
	 */

	callLineDescriptorTable[callLineNumber].controlChannelDescriptorp =
	  controlChannelDescriptorp;

	/*
	 * Begin call establishment for the call line
	 */

	if ((*callLineDescriptorTable[callLineNumber].sendCallRequestFunction)
	    (&callLineDescriptorTable[callLineNumber]) == PPTP_SUCCESS) {
	  DBG_LEAVE();
	  return PPTP_SUCCESS;
	}
	else {
	  ASSERT(!assertOnFatalError);
	  DBG_LEAVE();
	  return PPTP_FAILURE;
	}
      }
    }
  }

  /*
   * There were not any more control channel seek descriptors pending
   * for this control channel so return PPTP_SUCCESS.  It is the
   * responsibility of the code that handles any call line disconnection
   * to initate a shutdown of the control channel if it is no longer in
   * used.
   */

  DBG_LEAVE();
  return PPTP_SUCCESS;

}


/*
 * PptpContinuePendingControlChannelSeekDescriptors()
 *
 * This routine will continue any seek descriptors that were waiting
 * for a particular control channel descriptor to establish
 * a control channel and was unable to do so with this control channel.
 * This can happen when one of the following occurs:
 *
 *   - the control channel remotely disconnects
 *
 *   - the control channel will not accept any more calls from us
 */

STATIC PptpReturnCode PptpContinuePendingControlChannelSeekDescriptors(
	PptpControlChannelDescriptor	*controlChannelDescriptorp
	)
{
  PptpControlChannelSeekIndex controlChannelSeekNumber;

  DBG_FUNC("PptpContinuePendingControlChannelSeekDescriptors", PPTP_PROTOCOL);

  DBG_ENTER();

  /*
   * Go through the control channel seek descriptor table and continue
   * the control channel seek for each descriptor that was pending on this
   * control channel descriptor.
   */

  for (controlChannelSeekNumber = 0;
       controlChannelSeekNumber < PPTP_MAX_CONTROL_CHANNEL_SEEK_NUMBER ;
       controlChannelSeekNumber++) {
    if (controlChannelSeekDescriptorTable[controlChannelSeekNumber].
	pendingControlChannelDescriptorp ==
	controlChannelDescriptorp) {

      /*
       * This control channel seek descriptor was pending on the
       * control channel descriptor so unbind it and continue
       * the control channel seek descriptor search.  If the
       * control channel seek descriptor search fails then
       * abort it
       */

      controlChannelSeekDescriptorTable[controlChannelSeekNumber].
	pendingControlChannelDescriptorp = NULL;
      if (PptpContinueControlChannelSeekDescriptorSearch(
		&controlChannelSeekDescriptorTable[controlChannelSeekNumber])
	  != PPTP_SUCCESS) {
	PptpAbortControlChannelSeekDescriptorSearch(
		&controlChannelSeekDescriptorTable[controlChannelSeekNumber]);
      }
      ASSERT(controlChannelSeekDescriptorTable[controlChannelSeekNumber].
	     pendingControlChannelDescriptorp !=
	     controlChannelDescriptorp);
    }
  }

  DBG_LEAVE();
  return PPTP_SUCCESS;

}


STATIC PptpReturnCode PptpAttemptControlChannelConnection(
	PptpControlChannelDescriptor	*controlChannelDescriptorp
	)
{
  DBG_FUNC("PptpAttemptControlChannelConnection", PPTP_PROTOCOL);

  DBG_ENTER();

  ASSERT(controlChannelDescriptorp->controlChannelState ==
	 PPTP_CONTROL_CHANNEL_ALLOCATED);

  /*
   * Call PptpNetControlConnectReq() to initiate a control channel
   * connection to the remote peer.
   */

  if (PptpNetControlConnectReq(&controlChannelDescriptorp->
			       remoteNetConfigEntry.
			       remoteControlChannelAddress,
			       &controlChannelDescriptorp->
			       controlChannelHandle) != PPTP_SUCCESS) {
    DBG_LEAVE();
    return PPTP_FAILURE;
  }

  /*
   * Move the control channel to the PPTP_CONTROL_CHANNEL_CONNECTING state
   * and mark it as being a locally initiated connection.
   */

  controlChannelDescriptorp->controlChannelState =
    PPTP_CONTROL_CHANNEL_CONNECTING;
  controlChannelDescriptorp->locallyInitiatedConnection = 1;

  DBG_LEAVE();
  return PPTP_SUCCESS;

}


/*
 * PptpCheckControlChannelSeekDescriptorUsage()
 *
 * This routine checks to see a control channel seek descriptor is
 * required any longer.
 */

STATIC PptpReturnCode PptpCheckControlChannelSeekDescriptorUsage(
	PptpControlChannelSeekDescriptor	*controlChannelSeekDescriptorp
	)
{
  PptpCallLineNumber callLineNumber;
  PptpControlChannelDescriptor *controlChannelDescriptorp;

  DBG_FUNC("PptpCheckControlChannelSeekDescriptorUsage", PPTP_PROTOCOL);

  DBG_ENTER();

  /*
   * Verify that the control channel seek descriptor is bound
   * to a control channel descriptor
   */

  if (!controlChannelSeekDescriptorp->pendingControlChannelDescriptorp) {
    ASSERT(!assertOnFatalError);
    DBG_LEAVE();
    return PPTP_FAILURE;
  }

  /*
   * Check to see if any other calls are using the seek descriptor,
   * if any are then return immediately since the seek descriptor is
   * still being used
   */

  for (callLineNumber = 0 ;
       callLineNumber < currentPptpConfig->numberLineDevices ;
       callLineNumber++) {
    if (callLineDescriptorTable[callLineNumber].
	controlChannelSeekDescriptorp ==
	controlChannelSeekDescriptorp) {
      DBG_LEAVE();
      return PPTP_SUCCESS;
    }
  }

  /*
   * No calls are using the control channel seek descriptor.
   * Unbind the control channel descriptor from the control
   * channel seek descriptor since no other calls are using
   * the control channel seek descriptor
   */

  controlChannelDescriptorp =
    controlChannelSeekDescriptorp->pendingControlChannelDescriptorp;
  controlChannelSeekDescriptorp->pendingControlChannelDescriptorp =
    NULL;

  /*
   * Check to see if the control channel that was used by the seek
   * descriptor is required anymore
   */

  PptpCheckControlChannelDescriptorUsage(controlChannelDescriptorp);

  /*
   * Abort the control channel seek descriptor search
   */

  PptpAbortControlChannelSeekDescriptorSearch(controlChannelSeekDescriptorp);

  DBG_LEAVE();
  return PPTP_SUCCESS;

}


/*
 * PptpCheckControlChannelDescriptorUsage()
 *
 * This routine checks to see a control channel is being used any longer.
 * If not and the control channel was locally initiated then it
 * starts shutting down the control channel
 */

STATIC PptpReturnCode PptpCheckControlChannelDescriptorUsage(
	PptpControlChannelDescriptor	*controlChannelDescriptorp
	)
{
  PptpControlChannelSeekIndex controlChannelSeekNumber;
  PptpCallLineNumber callLineNumber;

  DBG_FUNC("PptpCheckControlChannelDescriptorUsage", PPTP_PROTOCOL);

  DBG_ENTER();

  /*
   * First check to see if the control channel was locally initiated, if
   * so then simply return
   */

  if (!controlChannelDescriptorp->locallyInitiatedConnection) {
    DBG_LEAVE();
    return PPTP_SUCCESS;
  }

  /*
   * Search through the control channel seek descriptor table for
   * any seek descriptors that may be using this control channel, if any
   * are found then simply return since the control channel
   * is still being used
   */

  for (controlChannelSeekNumber = 0 ;
       controlChannelSeekNumber < PPTP_MAX_CONTROL_CHANNEL_SEEK_NUMBER ;
       controlChannelSeekNumber++) {
    if (controlChannelSeekDescriptorTable[controlChannelSeekNumber].
	pendingControlChannelDescriptorp == controlChannelDescriptorp) {
      ASSERT(controlChannelSeekDescriptorTable[controlChannelSeekNumber].
	     controlChannelSeekState != PPTP_SEEK_NO_STATE);
      DBG_LEAVE();
      return PPTP_SUCCESS;
    }
  }

  /*
   * Search through the call line table for any calls that may
   * be using this control channel, if any are found then simply
   * return since the control channel is still being used
   */

  for (callLineNumber = 0 ;
       callLineNumber < currentPptpConfig->numberLineDevices ;
       callLineNumber++) {
    if (callLineDescriptorTable[callLineNumber].controlChannelDescriptorp ==
	controlChannelDescriptorp) {
      ASSERT(callLineDescriptorTable[callLineNumber].callLineState !=
	      PPTP_CALL_LINE_NO_STATE);
      DBG_LEAVE();
      return PPTP_SUCCESS;
    }
  }

  /*
   * This detached call was the last one using the control channel
   * so initiate a local shutdown of the control channel
   */

  PptpShutdownControlChannelDescriptor(controlChannelDescriptorp);

  DBG_LEAVE();

}


/*
 * PptpControlChannelRemotelyDisconnected()
 *
 * This routine is called when a control channel has been
 * terminated by the remote control channel peer
 */

STATIC PptpReturnCode PptpControlChannelRemotelyDisconnected(
	PptpControlChannelDescriptor	*controlChannelDescriptorp
	)
{
  PptpCallLineNumber callLineNumber;
  PptpControlChannelSeekIndex controlChannelSeekNumber;

  DBG_FUNC("PptpControlChannelRemotelyDisconnected", PPTP_PROTOCOL);

  DBG_ENTER();

  /*
   * If any calls were using a seek descriptor that was bound to this
   * control channel then detach the call from its control channel descriptor
   * so its search can continue, otherwise the call will be treated as
   * if it was connected
   */

  for (callLineNumber = 0 ;
       callLineNumber < currentPptpConfig->numberLineDevices ;
       callLineNumber++) {
    if ((callLineDescriptorTable[callLineNumber].
	 controlChannelSeekDescriptorp) &&
	(callLineDescriptorTable[callLineNumber].controlChannelDescriptorp ==
	 controlChannelDescriptorp)) {
      ASSERT(callLineDescriptorTable[callLineNumber].
	     controlChannelSeekDescriptorp->pendingControlChannelDescriptorp ==
	     controlChannelDescriptorp);
      callLineDescriptorTable[callLineNumber].controlChannelDescriptorp = NULL;
    }
  }

  /*
   * Notify any pending seek descriptors that were using this control
   * channel so they will attempt to use a new control channel
   */

  PptpContinuePendingControlChannelSeekDescriptors(controlChannelDescriptorp);

  /*
   * Any calls remaining that are using this control channel need to
   * be informed that their control channel has been lost which is
   * equivalent to a remote disconnect for the call.
   */

  for (callLineNumber = 0 ;
       callLineNumber < currentPptpConfig->numberLineDevices ;
       callLineNumber++) {
    if (callLineDescriptorTable[callLineNumber].controlChannelDescriptorp ==
	controlChannelDescriptorp) {
      PptpReturnCode status;

      /*
       * The call line no longer is using the control channel so
       * sever its connection to it
       */

      callLineDescriptorTable[callLineNumber].controlChannelDescriptorp =
	NULL;

      /*
       * If there was a flow control timer scheduled for the call
       * then unschedule it
       */

	if (callLineDescriptorTable[callLineNumber].flowControlTimer.timerScheduled) {
		PptpUnscheduleInternalTimer(&callLineDescriptorTable[callLineNumber].flowControlTimer);
      }

	  if (callLineDescriptorTable[callLineNumber].delayedAckTimer.timerScheduled) {
		  PptpUnscheduleInternalTimer(&callLineDescriptorTable[callLineNumber].delayedAckTimer);
	  }

#if defined(PPTP_SERVER)
      status =
	PptpServerCallLineLostControlChannel(&callLineDescriptorTable
					     [callLineNumber]);
#else
      status =
	PptpClientCallLineLostControlChannel(&callLineDescriptorTable
					     [callLineNumber]);
#endif /* defined(PPTP_SERVER) */

      ASSERT(status == PPTP_SUCCESS);
    }
  }

  /*
   * At this point no more seek descriptors or call lines should
   * be using this control channel.  For debugging purposes first
   * verify that no seek descriptors or call lines are using the
   * control channel.
   */

  for (controlChannelSeekNumber = 0 ;
       controlChannelSeekNumber < PPTP_MAX_CONTROL_CHANNEL_SEEK_NUMBER ;
       controlChannelSeekNumber++) {
    ASSERT(controlChannelSeekDescriptorTable[controlChannelSeekNumber].
	   pendingControlChannelDescriptorp !=
	   controlChannelDescriptorp);
  }

  for (callLineNumber = 0 ;
       callLineNumber < currentPptpConfig->numberLineDevices ;
       callLineNumber++) {
    ASSERT(callLineDescriptorTable[callLineNumber].
	   controlChannelDescriptorp !=
	   controlChannelDescriptorp);
  }

  /*
   * Call PptpTerminateInactivityIdleProcessing() to terminate any
   * inactivity idle timers
   */

  PptpTerminateInactivityIdleProcessing(controlChannelDescriptorp);

  DBG_LEAVE();
  return PPTP_SUCCESS;

}


/*
 * PptpShutdownControlChannelDescriptor()
 *
 * This routine is called to shutdown a locally initiated control channel
 * after all calls and seek descriptors have been detached from it
 */


STATIC PptpReturnCode PptpShutdownControlChannelDescriptor(
	PptpControlChannelDescriptor	*controlChannelDescriptorp
	)
{
  PptpStopSessionRequestMsg stopSessionRequestMsg;

  DBG_FUNC("PptpShutdownControlChannel", PPTP_PROTOCOL);

  DBG_ENTER();

  DBG_TRACE("Shutting down control channel\n");

  /*
   * Verify that we are initating the shutdown for a locally initiated
   * control channel
   */

  if (!controlChannelDescriptorp->locallyInitiatedConnection) {
    ASSERT(!assertOnFatalError);
    DBG_LEAVE();
    return PPTP_FAILURE;
  }

  /*
   * Check the control channel state for validity.  How the control
   * channel is shutdown depends on what state it is in
   */

  switch (controlChannelDescriptorp->controlChannelState) {

  case PPTP_CONTROL_CHANNEL_CONNECTING:
  case PPTP_CONTROL_CHANNEL_STARTING_SESSION:

    /*
     * Call PptpTerminateInactivityIdleProcessing() to terminate any
     * inactivity idle timers
     */

    PptpTerminateInactivityIdleProcessing(controlChannelDescriptorp);

    /*
     * The control channel is being shutdown before it was fully established
     * so disconnect the network connection.  First change the
     * control channel state to PPTP_CONTROL_CHANNEL_DISCONNECTING
     */

    controlChannelDescriptorp->controlChannelState =
      PPTP_CONTROL_CHANNEL_DISCONNECTING;

    /*
     * Call PptpNetControlDisconnectReq() to disconnect the control
     * channel
     */

    if (PptpNetControlDisconnectReq(controlChannelDescriptorp->
				    controlChannelHandle, 0) != PPTP_SUCCESS) {
      ASSERT(!assertOnFatalError);
      DBG_LEAVE();
      return PPTP_FAILURE;
    }

    break;

  case PPTP_CONTROL_CHANNEL_ESTABLISHED:

    /*
     * The control channel is being shutdown after it was already
     * established so start shutting it down by sending a
     * PPTP_STOP_SESSION_REQUEST message to the remote end
     */

    PptpBuildControlHeader(&stopSessionRequestMsg.header,
			   PPTP_STOP_SESSION_REQUEST,
			   sizeof(stopSessionRequestMsg));
    stopSessionRequestMsg.reason = PPTP_STOP_LOCAL_SHUTDOWN;
    stopSessionRequestMsg.reserved1 = 0;
    stopSessionRequestMsg.reserved2 = 0;

    if (PptpSendControlMessage(controlChannelDescriptorp,
			       (PptpControlPacket *) &stopSessionRequestMsg) ==
	PPTP_FAILURE) {
      ASSERT(!assertOnFatalError);
      DBG_LEAVE();
      return PPTP_FAILURE;
    }

    /*
     * Move the control session to the PPTP_CONTROL_CHANNEL_STOPPING_SESSION
     * state
     */

    controlChannelDescriptorp->controlChannelState =
      PPTP_CONTROL_CHANNEL_STOPPING_SESSION;

    break;

  default:
    ASSERT(!assertOnFatalError);
    DBG_LEAVE();
    return PPTP_FAILURE;
  }

  DBG_LEAVE();
  return PPTP_SUCCESS;

}


STATIC PptpReturnCode PptpControlChannelAddressMatch(
	PptpRemoteNetConfigEntry *firstRemoteNetConfigEntry,
	PptpRemoteNetConfigEntry *secondRemoteNetConfigEntry
	)
{
  PptpReturnCode status;

  if (!memcmp(&firstRemoteNetConfigEntry->
	      remoteControlChannelAddress.sin_addr.s_addr,
	      &secondRemoteNetConfigEntry->
	      remoteControlChannelAddress.sin_addr.s_addr,
	      sizeof(firstRemoteNetConfigEntry->
		     remoteControlChannelAddress.sin_addr.s_addr)) &&
      (firstRemoteNetConfigEntry->remoteProtocolNumber ==
       secondRemoteNetConfigEntry->remoteProtocolNumber)) {
    status = PPTP_SUCCESS;
  }
  else {
    status = PPTP_FAILURE;
  }

  return status;

}


STATIC VOID PptpSetControlChannelAddress(
	PptpRemoteNetConfigEntry	*destRemoteNetConfigEntry,
	PptpRemoteNetConfigEntry	*srcRemoteNetConfigEntry
	)
{
  DBG_FUNC("PptpSetControlChannelAddress", PPTP_PROTOCOL);

  DBG_ENTER();

  ASSERT(srcRemoteNetConfigEntry->entryInUse);

  destRemoteNetConfigEntry->remoteControlChannelAddress.sin_addr.s_addr =
    srcRemoteNetConfigEntry->remoteControlChannelAddress.sin_addr.s_addr;
  destRemoteNetConfigEntry->remoteControlChannelAddress.sin_port =
    srcRemoteNetConfigEntry->remoteControlChannelAddress.sin_port;
  destRemoteNetConfigEntry->remoteProtocolNumber =
    srcRemoteNetConfigEntry->remoteProtocolNumber;

  DBG_LEAVE();

}


PptpReturnCode PptpAssignNewCallLineNumber(
	PptpCallLineNumber *newcallLineNumber
	)
{
  PptpCallLineNumber callLineNumber;

  DBG_FUNC("PptpAssignNewCallLineNumber", PPTP_PROTOCOL);

  DBG_ENTER();

  /*
   * Search the callLineDescriptorTable for an unused call line.
   * Set the devices state to PPTP_CALL_LINE_NO_STATES.
   */

  for (callLineNumber = 0 ;
       callLineNumber < currentPptpConfig->numberLineDevices ;
       callLineNumber++) {
    if (callLineDescriptorTable[callLineNumber].callLineState ==
	PPTP_CALL_LINE_NO_STATE) {

      /*
       * Initialize the call line descriptor entry
       */

#if defined(PPTP_SYMMETRIC_SERVER)
    callLineDescriptorTable[callLineNumber].symmetricCallReceived = 0;
#endif /* defined(PPTP_SYMMETRIC_SERVER) */

      callLineDescriptorTable[callLineNumber].callLineState = PPTP_CALL_LINE_ALLOCATED;
      callLineDescriptorTable[callLineNumber].controlChannelDescriptorp = NULL;
      callLineDescriptorTable[callLineNumber].sendCallRequestFunction = NULL;
      callLineDescriptorTable[callLineNumber].localCallLineNumber = callLineNumber;
      callLineDescriptorTable[callLineNumber].remoteCallLineNumber = PPTP_UNUSED_CALL_LINE_NUMBER;
      callLineDescriptorTable[callLineNumber].recvDataPacketHeadIndex = 0;
      callLineDescriptorTable[callLineNumber].recvDataPacketPassedOnIndex = 0;
      callLineDescriptorTable[callLineNumber].recvDataPacketTailIndex = 0;
      callLineDescriptorTable[callLineNumber].lastOutputSeqNumber = 0;
      callLineDescriptorTable[callLineNumber].highestAckedOutputSeqNumber = 0;
      callLineDescriptorTable[callLineNumber].highestInputSeqNumber = 0;
      callLineDescriptorTable[callLineNumber].highestProcessedInputSeqNumber = 0;
      callLineDescriptorTable[callLineNumber].highestAckedInputSeqNumber = 0;
      callLineDescriptorTable[callLineNumber].numberDiscardedRecvPackets = 0;
      callLineDescriptorTable[callLineNumber].localDataChannelFlowControlOn = 0;
      callLineDescriptorTable[callLineNumber].remoteDataChannelFlowControlOn = 0;

      *newcallLineNumber = callLineNumber;
      DBG_LEAVE();
      return PPTP_SUCCESS;
    }
  }

  /*
   * An usused device entry could not be found, return PPTP_FAILURE.
   */

  DBG_LEAVE();
  return PPTP_FAILURE;

}


PptpReturnCode PptpCallLineNumberToCallLineDescriptorp(
	PptpCallLineNumber	callLineNumber,
	PptpCallLineDescriptor	**callLineDescriptorpp
	)
{

  DBG_FUNC("PptpCallLineNumberToCalLineDescriptorp", PPTP_PROTOCOL);

  DBG_ENTER();

  if ((callLineNumber >= 0 && callLineNumber < currentPptpConfig->numberLineDevices) &&
      (callLineDescriptorTable[callLineNumber].callLineState !=
       PPTP_CALL_LINE_NO_STATE)) {
    if (callLineDescriptorpp != NULL) {
      *callLineDescriptorpp = &callLineDescriptorTable[callLineNumber];
    }
    DBG_LEAVE();
    return PPTP_SUCCESS;
  }

  DBG_LEAVE();
  return PPTP_FAILURE;

}


PptpReturnCode PptpRemoteCallLineNumberToCallLineDescriptorp(
	PptpCallLineNumber				remoteCallLineNumber,
	PptpControlChannelDescriptor	*controlChannelDescriptorp,
	PptpCallLineDescriptor			**callLineDescriptorpp
	)
{
	PptpCallLineNumber localCallLineNumber;

	DBG_FUNC("PptpRemoteCallLineNumberToCallLineDescriptorp", PPTP_PROTOCOL);
	DBG_ENTER();

	for (localCallLineNumber = 0 ;
		localCallLineNumber < currentPptpConfig->numberLineDevices;
		localCallLineNumber++) {
		if ((callLineDescriptorTable[localCallLineNumber].callLineState !=
			PPTP_CALL_LINE_NO_STATE) &&
			(callLineDescriptorTable[localCallLineNumber].remoteCallLineNumber ==
			remoteCallLineNumber) &&
			(callLineDescriptorTable[localCallLineNumber].
			controlChannelDescriptorp == controlChannelDescriptorp)) {
			if (callLineDescriptorpp != NULL) {
				*callLineDescriptorpp =
				&callLineDescriptorTable[localCallLineNumber];
			}
			DBG_LEAVE();
			return PPTP_SUCCESS;
		}
	}

	DBG_LEAVE();
	return PPTP_FAILURE;
}


PptpReturnCode PptpCallDeviceNumberToCallLineNumber(
	PptpCallDeviceNumber	callDeviceNumber,
	PptpCallLineNumber	*callLineNumberp
	)
{
  PptpCallLineNumber callLineNumber;

  DBG_FUNC("PptpCallDeviceNumberToCallLineNumber", PPTP_PROTOCOL);

  DBG_ENTER();

  for (callLineNumber = 0 ;
       callLineNumber < currentPptpConfig->numberLineDevices ;
       callLineNumber++) {
    if ((callLineDescriptorTable[callLineNumber].callLineState !=
	 PPTP_CALL_LINE_NO_STATE) &&
	(callLineDescriptorTable[callLineNumber].callDeviceNumber ==
	 callDeviceNumber)) {
      *callLineNumberp = callLineNumber;
      DBG_LEAVE();
      return PPTP_SUCCESS;
    }
  }

  DBG_LEAVE();
  return PPTP_FAILURE;

}

PptpReturnCode PptpCallDeviceNumberToCallLineDescriptorp(
	PptpCallDeviceNumber	callDeviceNumber,
	PptpCallLineDescriptor	**callLineDescriptorp
	)
{
  PptpCallLineNumber callLineNumber;

  DBG_FUNC("PptpCallDeviceNumberToCallLineDescriptorp", PPTP_PROTOCOL);

  DBG_ENTER();

  if (PptpCallDeviceNumberToCallLineNumber(callDeviceNumber,
	  &callLineNumber) != PPTP_SUCCESS) {
	  DBG_LEAVE();
	  return PPTP_FAILURE;
  }

  *callLineDescriptorp = &callLineDescriptorTable[callLineNumber];

  DBG_LEAVE();
  return PPTP_FAILURE;

}


PptpReturnCode PptpLookupControlChannelDescriptorByHandle(
	PptpEndpointHandle		endpointHandle,
	PptpControlChannelDescriptor	**controlChannelDescriptorpp
	)
{
  PptpControlChannelNumber controlChannelNumber;

  DBG_FUNC("PptpLookupControlChannelNumberByHandle", PPTP_PROTOCOL);

  DBG_ENTER();

  /*
   * Search through controlChannelDescriptorTable for an entry that
   * has a controlChannelHandle that matches the endpointHandle.
   * This could be somewhat inefficient, but control message processing
   * should not be a bottleneck.
   */

  for (controlChannelNumber = 0;
       controlChannelNumber < currentPptpConfig->numberLineDevices ;
       controlChannelNumber++) {
    if (controlChannelDescriptorTable[controlChannelNumber].
	controlChannelHandle == endpointHandle) {
      *controlChannelDescriptorpp =
	&controlChannelDescriptorTable[controlChannelNumber];
      DBG_LEAVE();
      return PPTP_SUCCESS;
    }
  }

  DBG_LEAVE();
  return PPTP_FAILURE;

}


PptpReturnCode PptpLookupControlChannelNumberByAddr(
	PptpSockaddrIn			*remoteAddr,
	PptpProtocolNumber		remoteProtocolNumber,
	PptpControlChannelNumber	*controlChannelNumberp
	)
{
  PptpControlChannelNumber controlChannelNumber;

  DBG_FUNC("PptpLookupControlChannelNumberByAddr", PPTP_PROTOCOL);

  DBG_ENTER();


  /*
   * Search through controlChannelDescriptorTable for an entry that
   * has a remote address that matches the endpointHandle.
   */

  for (controlChannelNumber = 0;
       controlChannelNumber < currentPptpConfig->numberLineDevices ;
       controlChannelNumber++) {
    if ((controlChannelDescriptorTable[controlChannelNumber].
	 controlChannelState != PPTP_CONTROL_CHANNEL_NO_STATE) &&
	(controlChannelDescriptorTable[controlChannelNumber].
	 remoteNetConfigEntry.remoteControlChannelAddress.sin_addr.s_addr ==
	 remoteAddr->sin_addr.s_addr) &&
	(controlChannelDescriptorTable[controlChannelNumber].
	 remoteNetConfigEntry.remoteProtocolNumber ==
	 remoteProtocolNumber)) {
      *controlChannelNumberp = controlChannelNumber;
      DBG_LEAVE();
      return PPTP_SUCCESS;
    }
  }

  DBG_LEAVE();
  return PPTP_FAILURE;

}


VOID PptpBuildControlHeader(
	PptpControlMessageHeader	*headerp,
	PptpControlMessageType		messageType,
	USHORT				packetLength
	)
{

  DBG_FUNC("PptpBuildControlHeader", PPTP_PROTOCOL);

  DBG_ENTER();

  headerp->packetLength = htons(packetLength);
  headerp->packetType = htons(PPTP_CONTROL_PACKET);
  headerp->magicCookie = htonl(PPTP_MAGIC_COOKIE);
  headerp->messageType = htons((USHORT) messageType);
  headerp->reserved = 0;

  DBG_LEAVE();
}


PptpReturnCode PptpSendControlMessage(
	PptpControlChannelDescriptor	*controlChannelDescriptorp,
	PptpControlPacket		*controlPacket
	)
{
  ULONG dataLengthSent;

  DBG_FUNC("PptpSendControlMessage", PPTP_PROTOCOL);

  DBG_ENTER();


  ASSERT(ntohl(controlPacket->header.magicCookie) == PPTP_MAGIC_COOKIE);

  PptpTraceControlPacket(controlPacket, 0);

  if (PptpNetControlSendDataReq(controlChannelDescriptorp->
				controlChannelHandle,
				(UCHAR *) controlPacket,
				ntohs(controlPacket->header.packetLength),
				&dataLengthSent) != PPTP_SUCCESS) {
    DBG_LEAVE();
    return PPTP_FAILURE;
  }
  if (dataLengthSent != (ULONG) ntohs(controlPacket->header.packetLength)) {
    /*
     * TODO - write code to handle code in case socket won't eat all
     * the data if we intend on handling this in the future.
     */
    DBG_LEAVE();
    ASSERT(0);
    return PPTP_FAILURE;
  }

  DBG_LEAVE();
  return PPTP_SUCCESS;

}


STATIC PptpReturnCode PptpProtocolControlInput(
	PptpControlChannelDescriptor	*controlChannelDescriptorp,
	PptpControlPacket		*inputPacketp
	)
{
  ULONG desiredMessageSize;
  PptpReturnCode status;

  DBG_FUNC("PptpProtocolControlInput", PPTP_PROTOCOL);

  DBG_ENTER();


  PptpTraceControlPacket(inputPacketp, 1);

  /*
   * Convert the message header to host byte order
   */

  inputPacketp->header.packetType = ntohs(inputPacketp->header.packetType);
  inputPacketp->header.packetLength = ntohs(inputPacketp->header.packetLength);
  inputPacketp->header.magicCookie= ntohl(inputPacketp->header.magicCookie);
  inputPacketp->header.messageType = ntohs(inputPacketp->header.messageType);

  ASSERT(inputPacketp->header.packetLength >
	 sizeof(PptpControlMessageHeader));

  /*
   * Verify that the control channel is in a proper state to receive data
   * and check for general messages that should be received for a
   * particular state
   */

  switch (controlChannelDescriptorp->controlChannelState) {

  case PPTP_CONTROL_CHANNEL_STARTING_SESSION:

    switch (inputPacketp->header.messageType) {

    case PPTP_START_SESSION_REQUEST:
    case PPTP_START_SESSION_REPLY:
    case PPTP_ECHO_REQUEST:
    case PPTP_ECHO_REPLY:
      break;

    default:
      /* TODO - send an error message to the remote side */
      ASSERT(!assertOnProtocolError);
      DBG_LEAVE();
      return PPTP_FAILURE;
    }
    break;

  case PPTP_CONTROL_CHANNEL_ESTABLISHED:
    break;

  case PPTP_CONTROL_CHANNEL_STOPPING_SESSION:
    switch (inputPacketp->header.messageType) {

    case PPTP_STOP_SESSION_REPLY:
	case PPTP_ECHO_REQUEST:
	case PPTP_ECHO_REPLY:
		break;

    default:
      /* TODO - send an error message to the remote side */
      ASSERT(!assertOnProtocolError);
      DBG_LEAVE();
      return PPTP_FAILURE;
    }
    break;

  case PPTP_CONTROL_CHANNEL_DISCONNECTING:
  case PPTP_CONTROL_CHANNEL_NO_STATE:

    /*
     * It is possible to receive data during a disconnection, which should
     * just be discarded
     */

    DBG_LEAVE();
    return PPTP_SUCCESS;


  default:
    ASSERT(!assertOnFatalError);
    DBG_LEAVE();
    return PPTP_FAILURE;

  }

  switch (inputPacketp->header.messageType) {

  case PPTP_START_SESSION_REQUEST: {
    PptpStartSessionReplyMsg startSessionReply;

    /*
     * Validate the message size
     */

    desiredMessageSize = sizeof(PptpStartSessionRequestMsg);
    if (inputPacketp->header.packetLength != desiredMessageSize) {
      /* TODO - send an error message to the remote side. */
      ASSERT(!assertOnProtocolError);
      DBG_LEAVE();
      return PPTP_FAILURE;
    }

    /*
     * Make sure that the connection was started remotely, if not
     * then send a general error to the remote side.
     */

    if (controlChannelDescriptorp->locallyInitiatedConnection) {
      /* TODO - send an error message to the remote side. */
      ASSERT(!assertOnProtocolError);
      DBG_LEAVE();
      return PPTP_FAILURE;
    }

    /*
     * Send a PPTP_START_SESSION_REPLY message to the remote side.
     */

    PptpBuildControlHeader(&startSessionReply.header,
			   PPTP_START_SESSION_REPLY,
			   sizeof(startSessionReply));
    startSessionReply.protocolVersion = htons(PPTP_PROTOCOL_VERSION);
    startSessionReply.resultCode = PPTP_START_OK;
    startSessionReply.generalErrorCode = PPTP_ERROR_CODE_NONE;
    strncpy(startSessionReply.fepInfo.hostName,
	    currentPptpConfig->netConfig.localNetConfig.fepInfo.hostName,
	    MAX_HOSTNAME_LENGTH);
    startSessionReply.fepInfo.framingCapability =
	ntohl(PPTP_FRAME_CAP_ASYNC);
    startSessionReply.fepInfo.bearerCapability =
	ntohl(PPTP_BEARER_CAP_ANALOG);
    startSessionReply.fepInfo.maxChannels =
      htons(currentPptpConfig->netConfig.localNetConfig.fepInfo.maxChannels);
    startSessionReply.fepInfo.firmwareRevision =
      htons(currentPptpConfig->netConfig.localNetConfig.
	    fepInfo.firmwareRevision);
    strncpy(startSessionReply.fepInfo.vendorString,
	    currentPptpConfig->netConfig.localNetConfig.fepInfo.vendorString,
	    MAX_VENDOR_STRING_LENGTH);
    if (PptpSendControlMessage(controlChannelDescriptorp,
			       (PptpControlPacket *) &startSessionReply) !=
	PPTP_SUCCESS) {
      DBG_LEAVE();
      return PPTP_FAILURE;
    }

    /*
     * Enter the PPTP_CONTROL_CHANNEL_ESTABLISHED state.
     */

    controlChannelDescriptorp->controlChannelState =
      PPTP_CONTROL_CHANNEL_ESTABLISHED;

    /*
     * Continue and pending calls that were waiting on this control channel
     * to becomre established
     */

    PptpContinuePendingCallLineDescriptors(controlChannelDescriptorp);

    break;
  }

  case PPTP_START_SESSION_REPLY:

    /*
     * Validate the message size
     */

    desiredMessageSize = sizeof(PptpStartSessionReplyMsg);
    if (inputPacketp->header.packetLength != desiredMessageSize) {
      /* TODO - send an error message to the remote side. */
      ASSERT(!assertOnProtocolError);
      DBG_LEAVE();
      return PPTP_FAILURE;
    }

    /*
     * Make sure that the connection was started locally, if not
     * then send a general error to the remote side
     */

    if (!controlChannelDescriptorp->locallyInitiatedConnection) {
      /* TODO - send an error message to the remote side. */
      ASSERT(!assertOnProtocolError);
      DBG_LEAVE();
      return PPTP_FAILURE;
    }

    /*
     * Enter the PPTP_CONTROL_CHANNEL_ESTABLISHED state
     */

    controlChannelDescriptorp->controlChannelState =
      PPTP_CONTROL_CHANNEL_ESTABLISHED;

    /*
     * Continue and pending calls that were waiting on this control channel
     * to become established
     */

    PptpContinuePendingCallLineDescriptors(controlChannelDescriptorp);

    break;

  case PPTP_STOP_SESSION_REQUEST: {
    PptpStopSessionReplyMsg stopSessionReplyMsg;

    /*
     * Validate the message size
     */

    desiredMessageSize = sizeof(PptpStopSessionRequestMsg);
    if (inputPacketp->header.packetLength != desiredMessageSize) {
      /* TODO - send an error message to the remote side. */
      ASSERT(!assertOnProtocolError);
      DBG_LEAVE();
      return PPTP_FAILURE;
    }

    /*
     * Make sure that the connection was started remotely, if not
     * then send a general error to the remote side.
     */

    if (controlChannelDescriptorp->locallyInitiatedConnection) {
      /* TODO - send an error message to the remote side. */
      ASSERT(!assertOnProtocolError);
      DBG_LEAVE();
      return PPTP_FAILURE;
    }

    /*
     * Enter the PPTP_CONTROL_CHANNEL_STOPPING_SESSION state
     */

    controlChannelDescriptorp->controlChannelState =
      PPTP_CONTROL_CHANNEL_STOPPING_SESSION;

    /*
     * Send a PPTP_STOP_SESSION_REPLY message to the remote side
     */

    PptpBuildControlHeader(&stopSessionReplyMsg.header,
			   PPTP_STOP_SESSION_REPLY,
			   sizeof(stopSessionReplyMsg));
    stopSessionReplyMsg.resultCode = PPTP_STOP_OK;
    stopSessionReplyMsg.generalErrorCode = PPTP_ERROR_CODE_NONE;
    stopSessionReplyMsg.reserved1 = 0;

    if (PptpSendControlMessage(controlChannelDescriptorp,
			       (PptpControlPacket *) &stopSessionReplyMsg) !=
	PPTP_SUCCESS) {
      DBG_LEAVE();
      return PPTP_FAILURE;
    }

    /*
     * Change the control channel state to PPTP_CONTROL_CHANNEL_DISCONNECTING
     */

    controlChannelDescriptorp->controlChannelState =
      PPTP_CONTROL_CHANNEL_DISCONNECTING;

    /*
     * Inform all the call lines using this control channel that the
     * control channel has been remotely disconnected
     */

    PptpControlChannelRemotelyDisconnected(controlChannelDescriptorp);

    /*
     * Call PptpNetControlDisconnectReq() to disconnect the control
     * channel
     */

    if (PptpNetControlDisconnectReq(controlChannelDescriptorp->
				    controlChannelHandle, 0) !=
	PPTP_SUCCESS) {
      ASSERT(!assertOnFatalError);
      DBG_LEAVE();
      return PPTP_FAILURE;
    }

    break;
  }

  case PPTP_STOP_SESSION_REPLY:

    /*
     * Validate the message size
     */

    desiredMessageSize = sizeof(PptpStopSessionReplyMsg);
    if (inputPacketp->header.packetLength != desiredMessageSize) {
      /* TODO - send an error message to the remote side. */
      ASSERT(!assertOnProtocolError);
      DBG_LEAVE();
      return PPTP_FAILURE;
    }

    /*
     * Make sure that the connection was started locally, if not
     * then send a general error to the remote side.
     */

    if (!controlChannelDescriptorp->locallyInitiatedConnection) {
      /* TODO - send an error message to the remote side. */
      ASSERT(!assertOnProtocolError);
      DBG_LEAVE();
      return PPTP_FAILURE;
    }

    /*
     * Call PptpTerminateInactivityIdleProcessing() to terminate any
     * inactivity idle timers
     */

    PptpTerminateInactivityIdleProcessing(controlChannelDescriptorp);

    /*
     * Change the control channel state to PPTP_CONTROL_CHANNEL_DISCONNECTING
     */

    controlChannelDescriptorp->controlChannelState =
      PPTP_CONTROL_CHANNEL_DISCONNECTING;

    /*
     * Call PptpNetControlDisconnectReq() to disconnect the control
     * channel
     */

    if (PptpNetControlDisconnectReq(controlChannelDescriptorp->
				    controlChannelHandle, 0) != PPTP_SUCCESS) {
      ASSERT(!assertOnFatalError);
      DBG_LEAVE();
      return PPTP_FAILURE;
    }

    break;

  case PPTP_ECHO_REQUEST: {
//    PptpEchoRequestMsg *echoRequestp = &inputPacketp->echoRequest;
    PptpEchoReplyMsg echoReply;

    /*
     * Validate the message size
     */

    desiredMessageSize = sizeof(PptpEchoRequestMsg);
    if (inputPacketp->header.packetLength != desiredMessageSize) {
      /* TODO - send an error message to the remote side. */
      ASSERT(!assertOnProtocolError);
      DBG_LEAVE();
      return PPTP_FAILURE;
    }

#if DBG
	DbgPrint("RASPPTPE: <-EchoRequest %d cc: # %d s %d a %8.8x\n",
    ntohl(inputPacketp->echoRequest.identNumber),
    controlChannelDescriptorp->controlChannelIndex,
	controlChannelDescriptorp->controlChannelState,
    controlChannelDescriptorp->remoteNetConfigEntry.
    remoteControlChannelAddress.sin_addr.s_addr);
#endif

    /*
     * Send a PPTP_ECHO_REPLY message to the remote side.
     */

    PptpBuildControlHeader(&echoReply.header,
			   PPTP_ECHO_REPLY,
			   sizeof(echoReply));
    echoReply.identNumber = inputPacketp->echoRequest.identNumber;
    echoReply.resultCode = PPTP_ECHO_OK;
    echoReply.generalErrorCode = PPTP_ERROR_CODE_NONE;
    echoReply.reserved = 0;

    if (PptpSendControlMessage(controlChannelDescriptorp,
			       (PptpControlPacket *) &echoReply) !=
	PPTP_SUCCESS) {
		ASSERT(0);
      DBG_LEAVE();
      return PPTP_FAILURE;
    }
#if DBG
	DbgPrint("RASPPTPE: ->EchoReply %d cc: # %d s %d a %8.8x\n",
    ntohl(echoReply.identNumber),
    controlChannelDescriptorp->controlChannelIndex,
	controlChannelDescriptorp->controlChannelState,
    controlChannelDescriptorp->remoteNetConfigEntry.
    remoteControlChannelAddress.sin_addr.s_addr);
#endif

    break;
  }

	case PPTP_ECHO_REPLY:{
#if DBG
        DbgPrint("RASPPTPE: <-EchoReply %d cc: # %d s %d a %8.8x\n",
        ntohl(inputPacketp->echoReply.identNumber),
        controlChannelDescriptorp->controlChannelIndex,
        controlChannelDescriptorp->controlChannelState,
        controlChannelDescriptorp->remoteNetConfigEntry.
        remoteControlChannelAddress.sin_addr.s_addr);
#endif

	  break;
  }

#if defined(PPTP_SERVER) || defined(PPTP_SYMMETRIC_SERVER)
  case PPTP_OUT_CALL_REPLY:
    desiredMessageSize = sizeof(PptpOutCallReplyMsg);
    goto validate_server_packet;

  case PPTP_IN_CALL_REQUEST:
    desiredMessageSize = sizeof(PptpInCallRequestMsg);
    goto validate_server_packet;

  case PPTP_IN_CALL_CONNECTED:
    desiredMessageSize = sizeof(PptpInCallConnectedMsg);
    goto validate_server_packet;

  case PPTP_CALL_DISCONNECT_NOTIFY:
    desiredMessageSize = sizeof(PptpCallDisconnectNotifyMsg);
    goto validate_server_packet;

  case PPTP_WAN_ERROR_NOTIFY:
    desiredMessageSize = sizeof(PptpWanErrorNotifyMsg);
    goto validate_server_packet;

  validate_server_packet:

    /*
     * Validate the message size
     */

    if (inputPacketp->header.packetLength != desiredMessageSize) {
      /* TODO - send an error message to the remote side. */
      ASSERT(!assertOnProtocolError);
      DBG_LEAVE();
      return PPTP_FAILURE;
    }

    status = PptpServerControlProtocolInput(controlChannelDescriptorp,
					    inputPacketp);

    DBG_LEAVE();
    return status;

#endif /* defined(PPTP_SERVER) */

#if defined(PPTP_CLIENT) || defined(PPTP_SYMMETRIC_SERVER)
  case PPTP_OUT_CALL_REQUEST:
    desiredMessageSize = sizeof(PptpOutCallRequestMsg);
    goto validate_client_packet;

  case PPTP_IN_CALL_REPLY:
    desiredMessageSize = sizeof(PptpInCallReplyMsg);
    goto validate_client_packet;

  case PPTP_CALL_CLEAR_REQUEST:
    desiredMessageSize = sizeof(PptpCallClearRequestMsg);
    goto validate_client_packet;

  case PPTP_SET_LINK_INFO:
    desiredMessageSize = sizeof(PptpSetLinkInfoMsg);
    goto validate_client_packet;

  validate_client_packet:

    /*
     * Validate the message size
     */

    if (inputPacketp->header.packetLength != desiredMessageSize) {
      /* TODO - send an error message to the remote side. */
      ASSERT(!assertOnProtocolError);
      DBG_LEAVE();
      return PPTP_FAILURE;
    }

    status = PptpClientControlProtocolInput(controlChannelDescriptorp,
					    inputPacketp);

    DBG_LEAVE();
    return status;
#endif /* defined(PPTP_CLIENT) */

  default:
      /* TODO - send an error message to the remote side. */
      ASSERT(!assertOnProtocolError);
      DBG_LEAVE();
      return PPTP_FAILURE;

  }

  /*
   * The message has been successfully handled so reschedule the
   * control channel inactivity timer for the appropriate control
   * channel states
   */

  switch (controlChannelDescriptorp->controlChannelState) {
  case PPTP_CONTROL_CHANNEL_STARTING_SESSION:
  case PPTP_CONTROL_CHANNEL_ESTABLISHED:
  case PPTP_CONTROL_CHANNEL_STOPPING_SESSION:
    PptpRescheduleControlChannelInactivityTimer(controlChannelDescriptorp);
    break;
  }

  DBG_LEAVE();
  return PPTP_SUCCESS;

}


PptpReturnCode PptpDataChannelSend(
	PptpCallLineNumber	callLineNumber,
	UCHAR				*data,
	ULONG				dataLength,
	BOOL				*dataPacketDeliveredp,
	VOID				*completionArg
	)
{
  PptpCallLineDescriptor *callLineDescriptorp;
  PptpSeqNumber NextOutputSeqNumber;
  PptpDataDescriptor dataDescriptor;

  DBG_FUNC("PptpDataChannelSend", PPTP_PROTOCOL);

  DBG_ENTER();


  /*
   * Lookup the call information for the call represented by
   * callLineNumber
   */

  if (PptpCallLineNumberToCallLineDescriptorp(callLineNumber,
					      &callLineDescriptorp) != PPTP_SUCCESS) {
	ASSERT(!assertOnFatalError);
	*dataPacketDeliveredp = 0;
    DBG_LEAVE();
    return PPTP_FAILURE;
  }

  /*
   * Verify the call line descriptor is in the PPTP_CALL_LINE_CONNECTED
   * state
   */

  if (callLineDescriptorp->callLineState != PPTP_CALL_LINE_CONNECTED) {
    ASSERT(!assertOnFatalError);
    *dataPacketDeliveredp = 0;
    DBG_LEAVE();
    return PPTP_FAILURE;
  }

  /*
   * Check to see if the remote endpoint has any space left in its
   * packet window.  If so then send the packet, otherwise we need to
   * enable flow control to the remote endpoint
   */

	if ((LONG)(callLineDescriptorp->lastOutputSeqNumber -
		callLineDescriptorp->highestAckedOutputSeqNumber) >=
		callLineDescriptorp->remoteRxPacketWindow) {

    /*
     * Turn on flow control if it is enabled.  First determine what type
     * of flow control we are using.
     */

    switch (currentPptpConfig->dataChannelFlowControlType) {

    case PPTP_DATA_CHANNEL_FLOW_CONTROL_DISABLED:

      /*
       * If flow control is not enabled then don't do anything
       */
      DBG_TRACE("Output overflowing ack window since flow control disabled\n");
      break;

    case PPTP_DATA_CHANNEL_FLOW_CONTROL_STATIC_TIMEOUT:

      /*
       * Mark the packet as undeliverable
       */

      *dataPacketDeliveredp = 0;

      /*
       * For static timeout flow control we schedule a timer to
       * clear clear the window and resend when the timer expires
       * or a new packet is received
       */
      if (!callLineDescriptorp->flowControlTimer.timerScheduled) {
		  ASSERT(currentPptpConfig->dataChannelFlowControlStaticTimeoutSeconds >= 0);
		  if (PptpScheduleInternalTimer(
		      PptpDataChannelStaticTimeoutCallback,
		      callLineDescriptorp,
		      currentPptpConfig->
		      dataChannelFlowControlStaticTimeoutSeconds
		        * 1000,
		      &callLineDescriptorp->flowControlTimer) != PPTP_SUCCESS) {
			  ASSERT(!assertOnFatalError);
			  DBG_LEAVE();
			  return PPTP_FAILURE;
		  }
		  callLineDescriptorp->remoteDataChannelFlowControlOn = 1;
		  DBG_TRACE("static data channel timeout scheduled for call line = %d\n",
		  callLineDescriptorp->localCallLineNumber);
      }
      DBG_LEAVE();
      return PPTP_SUCCESS;

    case PPTP_DATA_CHANNEL_FLOW_CONTROL_SLIDING_WINDOW:
		ASSERT(!assertOnFatalError);
		*dataPacketDeliveredp = 0;
		DBG_LEAVE();
		return PPTP_FAILURE;

    default:
      ASSERT(!assertOnFatalError);
      *dataPacketDeliveredp = 0;
      DBG_LEAVE();
      return PPTP_FAILURE;
    }

  }

  /*
   * Increment the lastOutputSeqNumber field in callLineDescriptorp
   * while checking for wrap around
   */

  if (callLineDescriptorp->lastOutputSeqNumber == PPTP_MAX_SEQUENCE_NUMBER) {
    callLineDescriptorp->lastOutputSeqNumber = 0;
  }
  else {
    callLineDescriptorp->lastOutputSeqNumber++;
  }

  /*
   * Trace the data packet before sending it
   */

  PptpTraceDataPacket(callLineDescriptorp->
		      remoteCallLineNumber,
		      callLineDescriptorp->
		      lastOutputSeqNumber,
		      callLineDescriptorp->
		      highestProcessedInputSeqNumber,
		      data,
		      dataLength,
		      0);

  /*
   * Output the data packet
   */

   dataDescriptor.callLineNumber = callLineDescriptorp->remoteCallLineNumber;
   dataDescriptor.seqNumberPresent = TRUE;
   dataDescriptor.seqNumber = callLineDescriptorp->lastOutputSeqNumber;
   dataDescriptor.ackNumberPresent =
   ((LONG)(callLineDescriptorp->highestProcessedInputSeqNumber -
   callLineDescriptorp->highestAckedInputSeqNumber) > 0);
   dataDescriptor.ackNumber = callLineDescriptorp->highestProcessedInputSeqNumber;
   dataDescriptor.data = data;
   dataDescriptor.dataLength = dataLength;
   dataDescriptor.completionArg = completionArg;

  if (PptpNetDataChannelSendDataReq(dataChannelEndpointHandle,
	  &dataDescriptor,
	  &callLineDescriptorp->controlChannelDescriptorp->
	  remoteNetConfigEntry.remoteControlChannelAddress,
	  callLineDescriptorp->controlChannelDescriptorp->
	  remoteNetConfigEntry.remoteProtocolNumber) != PPTP_SUCCESS) {
    ASSERT(!assertOnFatalError);
    *dataPacketDeliveredp = 0;
    DBG_LEAVE();
    return PPTP_FAILURE;
  }

  //
  // Did this ack something higher than our current highest acked?
  //
  if (dataDescriptor.ackNumberPresent) {
	  callLineDescriptorp->highestAckedInputSeqNumber =
	  dataDescriptor.ackNumber;
  }

  *dataPacketDeliveredp = 1;
  DBG_LEAVE();
  return PPTP_SUCCESS;

}


PptpReturnCode PptpDataChannelRecvComplete(
	PptpCallLineNumber	callLineNumber,
	VOID			*completionArg
	)
{
  PptpCallLineDescriptor *callLineDescriptorp;

  DBG_FUNC("PptpDataChannelRecvComplete", PPTP_PROTOCOL);

  DBG_ENTER();

  /*
   * Call PptpNetDataChannelRecvDataCompleteReq() to inform the
   * network library that the received buffer is no longer required
   */

  if (PptpNetDataChannelRecvDataCompleteReq(completionArg) != PPTP_SUCCESS) {
    ASSERT(!assertOnFatalError);
  }

  /*
   * Lookup the call information for the call represented by
   * callLineNumber.  The call line may have been disconnected between
   * the time the receive packet was sent and this completion routine is
   * called so simply return if so.
   */

  if (PptpCallLineNumberToCallLineDescriptorp(callLineNumber,
					      &callLineDescriptorp) !=
      PPTP_SUCCESS) {
    DBG_LEAVE();
    return PPTP_SUCCESS;
  }

  /*
   * Verify the completionArg matches the argument stored in the
   * first packet in the pending data packet list to ensure that calls
   * to this function are not called out of order or dropped
   */

  if (callLineDescriptorp->recvDataPacketArray
      [callLineDescriptorp->recvDataPacketHeadIndex].completionArg !=
      completionArg) {
    ASSERT(!assertOnFatalError);
  }

	callLineDescriptorp->recvDataPacketArray
	[callLineDescriptorp->recvDataPacketHeadIndex].completionArg = NULL;

	//
	// Bump our highest processed sequence number
	//
	if ((LONG)(callLineDescriptorp->recvDataPacketArray
		[callLineDescriptorp->recvDataPacketHeadIndex].seqNumber -
		callLineDescriptorp->highestProcessedInputSeqNumber) > 0 ) {
		callLineDescriptorp->highestProcessedInputSeqNumber =
		callLineDescriptorp->recvDataPacketArray
		[callLineDescriptorp->recvDataPacketHeadIndex].seqNumber;
	}

  /*
   * Skip to the next entry in the receive data packet list by
   * incrementing the head index
   */

	if (++callLineDescriptorp->recvDataPacketHeadIndex ==
		PPTP_MAX_PACKET_WINDOW) {
		callLineDescriptorp->recvDataPacketHeadIndex = 0;
	}

  /*
   * Check to see if the call line is in the PPTP_CALL_LINE_CONNECTED state.
   * The call line may have been disconnected between the time the
   * receive packet was handed of and this completion routine is called
   * so simply return if so.
   */

  if (callLineDescriptorp->callLineState != PPTP_CALL_LINE_CONNECTED) {
    DBG_LEAVE();
    return PPTP_SUCCESS;
  }

  /*
   * If the input data channel was flow controlled then turn
   * off the input data channel flow control and call
   * PptpServicePendingPacketList to send any pending packets
   */

  if (callLineDescriptorp->localDataChannelFlowControlOn) {
	  callLineDescriptorp->localDataChannelFlowControlOn = 0;
    PptpServicePendingPacketList(callLineDescriptorp);
  }

    /*
     * Call PptpAckRemoteDataChannel to ack the remote data channel if
     * necessary
     */

    if ((LONG)(callLineDescriptorp->highestProcessedInputSeqNumber -
        callLineDescriptorp->highestAckedInputSeqNumber) >=
        callLineDescriptorp->localRxPacketWindow) {

        //
        // If we detect that our receive window is full (the sender is
        // flow controlled) we should ack immediately.
        // This will keep the sender from sending a window's worth of
        // data and then having to wait 250ms before it can send again.
        //
        PptpAckRemoteDataChannel(callLineDescriptorp);

    } else {
        if (!callLineDescriptorp->delayedAckTimer.timerScheduled) {
            if (PptpScheduleInternalTimer(PptpAckRemoteDataChannel,
                                          callLineDescriptorp,
                                          250,
                                          &callLineDescriptorp->delayedAckTimer)
                != PPTP_SUCCESS) {
                //
                // Could not get a timer scheduled so go ahead and ack
                //
                PptpAckRemoteDataChannel(callLineDescriptorp);
            }
        }
    }

    DBG_LEAVE();
    return PPTP_SUCCESS;
}


/*
 * PptpGetDebugStateInfo()
 *
 * This routine fills in a debug state information structure with
 * the current values contained in the PPTP data structures.  This
 * information is useful for debugging purposes
 */

PptpReturnCode PptpGetDebugStateInfo(
	PptpDebugStateInfo	*debugStateInfop
	)
{
  PptpCallLineNumber callLineNumber;
  PptpControlChannelIndex controlChannelIndex;
  PptpControlChannelSeekIndex controlChannelSeekIndex;

  DBG_FUNC("PptpGetDebugStateInfo", PPTP_PROTOCOL);
  DBG_ENTER();

  /*
   * Fill in the config structure
   */

	debugStateInfop->config = *currentPptpConfig;

  /*
   * Fill in the members of the callLineInfo structure
   */

  for (callLineNumber = 0 ;
       callLineNumber < currentPptpConfig->numberLineDevices ;
       callLineNumber++) {
    debugStateInfop->callLineInfo[callLineNumber].callLineState =
      callLineDescriptorTable[callLineNumber].callLineState;
    if (callLineDescriptorTable[callLineNumber].controlChannelDescriptorp) {
      debugStateInfop->callLineInfo[callLineNumber].controlChannelIndex =
	callLineDescriptorTable[callLineNumber].controlChannelDescriptorp->
	  controlChannelIndex;
    }
    else {
      debugStateInfop->callLineInfo[callLineNumber].controlChannelIndex =
	PPTP_ENTRY_NOT_IN_USE;
    }
    if (callLineDescriptorTable[callLineNumber].
	controlChannelSeekDescriptorp) {
      debugStateInfop->callLineInfo[callLineNumber].controlChannelSeekIndex =
	callLineDescriptorTable[callLineNumber].controlChannelSeekDescriptorp->
	  controlChannelSeekIndex;
    }
    else {
      debugStateInfop->callLineInfo[callLineNumber].controlChannelSeekIndex =
	PPTP_ENTRY_NOT_IN_USE;
    }
    debugStateInfop->callLineInfo[callLineNumber].remoteCallLineNumber =
      callLineDescriptorTable[callLineNumber].remoteCallLineNumber;
    debugStateInfop->callLineInfo[callLineNumber].callSerialNumber =
      callLineDescriptorTable[callLineNumber].callSerialNumber;
    debugStateInfop->callLineInfo[callLineNumber].physChannelID =
      callLineDescriptorTable[callLineNumber].physChannelID;
    debugStateInfop->callLineInfo[callLineNumber].connectBPS =
      callLineDescriptorTable[callLineNumber].connectBPS;
    debugStateInfop->callLineInfo[callLineNumber].callBearerType =
      callLineDescriptorTable[callLineNumber].callBearerType;
    debugStateInfop->callLineInfo[callLineNumber].framingType =
      callLineDescriptorTable[callLineNumber].framingType;
    debugStateInfop->callLineInfo[callLineNumber].localRxPacketWindow =
      callLineDescriptorTable[callLineNumber].localRxPacketWindow;
    debugStateInfop->callLineInfo[callLineNumber].remoteRxPacketWindow =
      callLineDescriptorTable[callLineNumber].remoteRxPacketWindow;
    debugStateInfop->callLineInfo[callLineNumber].remoteRxPacketWindowMax =
        callLineDescriptorTable[callLineNumber].remoteRxPacketWindowMax;
    debugStateInfop->callLineInfo[callLineNumber].localRxPacketProcDelay =
      callLineDescriptorTable[callLineNumber].localRxPacketProcDelay;
    debugStateInfop->callLineInfo[callLineNumber].remoteRxPacketProcDelay =
      callLineDescriptorTable[callLineNumber].remoteRxPacketProcDelay;
  }

  /*
   * Fill in the members of the controlChannelInfo structure
   */

  for (controlChannelIndex = 0 ;
       controlChannelIndex < currentPptpConfig->numberLineDevices ;
       controlChannelIndex++) {
    debugStateInfop->controlChannelInfo[controlChannelIndex].
      controlChannelState =
	controlChannelDescriptorTable[controlChannelIndex].
	  controlChannelState;
    debugStateInfop->controlChannelInfo[controlChannelIndex].
      locallyInitiatedConnection =
	controlChannelDescriptorTable[controlChannelIndex].
	  locallyInitiatedConnection;
    debugStateInfop->controlChannelInfo[controlChannelIndex].
      remoteNetConfigEntry =
	controlChannelDescriptorTable[controlChannelIndex].
	  remoteNetConfigEntry;
  }

  /*
   * Fill in the members of the controlChanelDebugInfo structure
   */

  for (controlChannelSeekIndex = 0 ;
       controlChannelSeekIndex < PPTP_MAX_CONTROL_CHANNEL_SEEK_NUMBER ;
       controlChannelSeekIndex++) {
    debugStateInfop->controlChannelSeekInfo[controlChannelSeekIndex].
      controlChannelSeekState =
	controlChannelSeekDescriptorTable[controlChannelSeekIndex].
	  controlChannelSeekState;
    if (controlChannelSeekDescriptorTable[controlChannelSeekIndex].
	pendingControlChannelDescriptorp) {
    debugStateInfop->controlChannelSeekInfo[controlChannelSeekIndex].
      controlChannelIndex =
	controlChannelSeekDescriptorTable[controlChannelSeekIndex].
	  pendingControlChannelDescriptorp->controlChannelIndex;
    }
    else {
      debugStateInfop->controlChannelSeekInfo[controlChannelSeekIndex].
	controlChannelIndex = PPTP_ENTRY_NOT_IN_USE;
    }
    debugStateInfop->controlChannelSeekInfo[controlChannelSeekIndex].
      currentAddressListIndex =
	controlChannelSeekDescriptorTable[controlChannelSeekIndex].
	  currentAddressListIndex;
    debugStateInfop->controlChannelSeekInfo[controlChannelSeekIndex].
      finalAddressListIndex =
	controlChannelSeekDescriptorTable[controlChannelSeekIndex].
	  finalAddressListIndex;
    debugStateInfop->controlChannelSeekInfo[controlChannelSeekIndex].
      lastEstablishedAddressListIndex =
	controlChannelSeekDescriptorTable[controlChannelSeekIndex].
	  lastEstablishedAddressListIndex;
  }

  DBG_LEAVE();
  return PPTP_SUCCESS;

}


STATIC VOID PptpServicePendingPacketList(
	PptpCallLineDescriptor		*callLineDescriptorp
	)
{
	PptpReturnCode status;
	BOOL packetDelivered;
	ULONG	recvDataPacketPassedOnIndex;
	PptpDataPacketDescriptor *dataPacketDescriptorp;

	DBG_FUNC("PptpServicePendingPacketList", PPTP_PROTOCOL);

	DBG_ENTER();

	/*
	 * Send the next packet pending if it will be accepted (no flow control)
	 */

	recvDataPacketPassedOnIndex =
	callLineDescriptorp->recvDataPacketPassedOnIndex;

	while (!callLineDescriptorp->localDataChannelFlowControlOn &&
		recvDataPacketPassedOnIndex !=
		callLineDescriptorp->recvDataPacketTailIndex) {

		dataPacketDescriptorp =
		&callLineDescriptorp->recvDataPacketArray[recvDataPacketPassedOnIndex];

		ASSERT(dataPacketDescriptorp->completionArg != NULL);

		/*
		 * Have the client or server process the received packet data
		 */

#if defined(PPTP_SERVER)
		status = PptpServerDataChannelRecv(callLineDescriptorp,
		                                   dataPacketDescriptorp->data,
										   dataPacketDescriptorp->dataLength,
										   &packetDelivered,
										   dataPacketDescriptorp->completionArg);
#else
		status = PptpClientDataChannelRecv(callLineDescriptorp,
		                                   dataPacketDescriptorp->data,
										   dataPacketDescriptorp->dataLength,
										   &packetDelivered,
										   dataPacketDescriptorp->completionArg);
#endif /* defined(PPTP_SERVER) */

		if (status == PPTP_SUCCESS && !packetDelivered) {
	
			/*
			 * The packet has been flow controlled so enable local data
			 * channel flow control and do not process any more packets.
			 * Packet processing will continue when the PptpDataChannelRecvComplete()
			 * routine is called to free up any packets pending in the
			 * client or server API routines
			 */
	
			callLineDescriptorp->localDataChannelFlowControlOn = 1;
			DBG_LEAVE();
			return;
		} else if (status != PPTP_SUCCESS) {
	
			/*
			 * The client our server would not accept the packet so
			 * release it and go onto the next packet.  For now, we do not
			 * handle this processing and will assert
			 */
	
			ASSERT(!packetDelivered);
			ASSERT(!assertOnFatalError);
			DBG_LEAVE();
			return;
		}
	
//		 if ((LONG)(callLineDescriptorp->recvDataPacketArray
//			 [recvDataPacketPassedOnIndex].seqNumber -
//			 callLineDescriptorp->highestProcessedInputSeqNumber) > 0 ) {
//			 callLineDescriptorp->highestProcessedInputSeqNumber =
//			 callLineDescriptorp->recvDataPacketArray
//			 [recvDataPacketPassedOnIndex].seqNumber;
//		 }

		/*
		 * Increment the recvDataPacketPassedOnIndex, checking for wraparound
		 */

		 if (++recvDataPacketPassedOnIndex ==
			 PPTP_MAX_PACKET_WINDOW) {
			 recvDataPacketPassedOnIndex = 0;
		 }

		 callLineDescriptorp->recvDataPacketPassedOnIndex =
		 recvDataPacketPassedOnIndex;
	}


	DBG_LEAVE();
}


STATIC VOID PptpAckRemoteDataChannel(
	PptpCallLineDescriptor		*callLineDescriptorp
	)
{
	PptpDataDescriptor	dataDescriptor;

  DBG_FUNC("PptpAckRemoteDataChannel", PPTP_PROTOCOL);

  DBG_ENTER();

  if (callLineDescriptorp->callLineState == PPTP_CALL_LINE_CONNECTED) {
	
	  /*
	   * If the max input sequence number for the data channel has not
	   * been sent to the remote end (we let it try to piggyback with output
	   * sent over the data channel first) then send it now with no data
	   */
	
	  if ((LONG)(callLineDescriptorp->highestProcessedInputSeqNumber -
          callLineDescriptorp->highestAckedInputSeqNumber) > 0) {
	
		PptpTraceDataPacket(callLineDescriptorp->
				   remoteCallLineNumber,
				   callLineDescriptorp->
				   lastOutputSeqNumber,
				   callLineDescriptorp->
				   highestProcessedInputSeqNumber,
				   NULL,
				   0,
				   0);
	
		dataDescriptor.callLineNumber = callLineDescriptorp->remoteCallLineNumber;
		dataDescriptor.seqNumberPresent = FALSE;
		dataDescriptor.seqNumber = callLineDescriptorp->lastOutputSeqNumber;
		dataDescriptor.ackNumberPresent = TRUE;
		dataDescriptor.ackNumber = callLineDescriptorp->highestProcessedInputSeqNumber;
		dataDescriptor.data = NULL;
		dataDescriptor.dataLength = 0;
		dataDescriptor.completionArg = NULL;

		if (PptpNetDataChannelSendDataReq(dataChannelEndpointHandle,
			&dataDescriptor,
			&callLineDescriptorp->controlChannelDescriptorp->
			remoteNetConfigEntry.remoteControlChannelAddress,
			callLineDescriptorp->controlChannelDescriptorp->
			remoteNetConfigEntry.remoteProtocolNumber) != PPTP_SUCCESS) {
		  ASSERT(!assertOnFatalError);
		}
	
		callLineDescriptorp->highestAckedInputSeqNumber =
		dataDescriptor.ackNumber;
	  }
  }


  DBG_LEAVE();
}


STATIC VOID PptpDataChannelStaticTimeoutCallback(
	PptpCallLineDescriptor		*callLineDescriptorp
	)
{
	PptpReturnCode Status;

	DBG_FUNC("PptpDataChannelStaticTimeoutCallback", PPTP_PROTOCOL);
	DBG_ENTER();
	DBG_TRACE("Static timer expired for call line = %d\n",
	          callLineDescriptorp->localCallLineNumber);

	ASSERT(currentPptpConfig->dataChannelFlowControlType ==
	       PPTP_DATA_CHANNEL_FLOW_CONTROL_STATIC_TIMEOUT);

	if (callLineDescriptorp->callLineState != PPTP_CALL_LINE_CONNECTED) {
		ASSERT(!assertOnFatalError);
		DBG_LEAVE();
		return;
	}

	if (callLineDescriptorp->remoteDataChannelFlowControlOn) {

		/*
		 * The data channel flow control type is a static timer which has
		 * expired, so clear the acknowledgement window
		 */
        //
        // Make it look like we have received ack's for all of the
        // send's we have made
        //
		callLineDescriptorp->highestAckedOutputSeqNumber =
		callLineDescriptorp->lastOutputSeqNumber;

        //
        // Reset the remote's receive window for a slow start
        //
        callLineDescriptorp->remoteRxPacketWindow = 1;

		callLineDescriptorp->remoteDataChannelFlowControlOn = 0;

#if defined(PPTP_SERVER)
		Status = PptpServerDataChannelSendAvailable(callLineDescriptorp);
#else
		Status = PptpClientDataChannelSendAvailable(callLineDescriptorp);
#endif /* defined(PPTP_SERVER) */

	}

	DBG_LEAVE();
}


STATIC PptpSeqNumber PptpSeqNumberDifference(
	PptpSeqNumber			baseSeqNumber,
	PptpSeqNumber			otherSeqNumber
	)
{
  DBG_FUNC("PptpSeqNumberDifference", PPTP_PROTOCOL);

  DBG_ENTER();

  if (baseSeqNumber >= otherSeqNumber) {
    DBG_LEAVE();
    return(baseSeqNumber - otherSeqNumber);
  }
  else {
    DBG_LEAVE();
    return(otherSeqNumber - baseSeqNumber + PPTP_MAX_SEQUENCE_NUMBER);
  }

}


STATIC PptpReturnCode PptpRescheduleControlChannelInactivityTimer(
	PptpControlChannelDescriptor	*controlChannelDescriptorp
	)
{
  DBG_FUNC("PptpRescheduleControlChannelInactivityTimer", PPTP_PROTOCOL);

  DBG_ENTER();

  /*
   * If there is a pending inactivity timer then cancel it
   */

  if (controlChannelDescriptorp->inactivityTimer.timerScheduled) {
    PptpUnscheduleInternalTimer(&controlChannelDescriptorp->inactivityTimer);
  }

  /*
   * Change the control channel inactivity state to
   * PPTP_CONTROL_CHANNEL_INACTIVITY_IDLE_STATE
   */

  controlChannelDescriptorp->controlChannelInactivityState =
    PPTP_CONTROL_CHANNEL_INACTIVITY_IDLE_STATE;

  /*
   * Schedule an idle inactivity timer
   */

  if (currentPptpConfig->inactivityIdleSeconds >= 0) {
    if (PptpScheduleInternalTimer(PptpInactivityTimerCallback,
				  controlChannelDescriptorp,
				  currentPptpConfig->inactivityIdleSeconds *
				  1000,
				  &controlChannelDescriptorp->
				  inactivityTimer) != PPTP_SUCCESS) {
      ASSERT(!assertOnFatalError);
      DBG_LEAVE();
      return PPTP_FAILURE;
    }
  }

  DBG_LEAVE();
  return PPTP_SUCCESS;

}


STATIC VOID PptpInactivityTimerCallback(
	PptpControlChannelDescriptor	*controlChannelDescriptorp
	)
{
  PptpEchoRequestMsg echoRequestMsg;

  DBG_FUNC("PptpInactivityTimerCallback", PPTP_PROTOCOL);

  DBG_ENTER();

  /*
   * Verify the control channel inactivity state
   */

  if (controlChannelDescriptorp->controlChannelInactivityState !=
      PPTP_CONTROL_CHANNEL_INACTIVITY_IDLE_STATE) {
    ASSERT(!assertOnFatalError);
    DBG_LEAVE();
    return;
  }

  /*
   * The inactivity timer has expired so send a PPTP_ECHO_REQUEST
   * message to the remote end
   */

	PptpBuildControlHeader(&echoRequestMsg.header,
	                       PPTP_ECHO_REQUEST,
						   sizeof(echoRequestMsg));

	controlChannelDescriptorp->echoIdentNumber++;

	echoRequestMsg.identNumber =
	htonl(controlChannelDescriptorp->echoIdentNumber);

	if (PptpSendControlMessage(controlChannelDescriptorp,
		(PptpControlPacket *) &echoRequestMsg) == PPTP_FAILURE) {
		ASSERT(!assertOnFatalError);
		DBG_LEAVE();
        return;
	}
#if DBG
	DbgPrint("RASPPTPE: ->EchoRequest %d cc: # %d s %d a %8.8x\n",
    controlChannelDescriptorp->echoIdentNumber,
    controlChannelDescriptorp->controlChannelIndex,
	controlChannelDescriptorp->controlChannelState,
    controlChannelDescriptorp->remoteNetConfigEntry.
    remoteControlChannelAddress.sin_addr.s_addr);
#endif

	/*
	 * Change the control channel inactivity state to
	 * PPTP_CONTROL_CHANNEL_INACTIVITY_WAITING_ECHO_RESPONSE_STATE
	 */

	controlChannelDescriptorp->controlChannelInactivityState =
	PPTP_CONTROL_CHANNEL_INACTIVITY_WAITING_ECHO_RESPONSE_STATE;

	/*
	 * Schedule an echo response timer
	 */

	if (currentPptpConfig->echoReplyTimeoutSeconds >= 0) {
		if (PptpScheduleInternalTimer(PptpEchoResponseTimerCallback,
			controlChannelDescriptorp,
			currentPptpConfig->echoReplyTimeoutSeconds * 1000,
			&controlChannelDescriptorp->inactivityTimer) != PPTP_SUCCESS) {
			ASSERT(!assertOnFatalError);
		}
	}


	DBG_LEAVE();
}


STATIC VOID PptpEchoResponseTimerCallback(
	PptpControlChannelDescriptor	*controlChannelDescriptorp
	)
{
  DBG_FUNC("PptpInactivityTimerCallback", PPTP_PROTOCOL);

  DBG_ENTER();

  /*
   * Verify the control channel inactivity state
   */

  if (controlChannelDescriptorp->controlChannelInactivityState !=
      PPTP_CONTROL_CHANNEL_INACTIVITY_WAITING_ECHO_RESPONSE_STATE) {
    ASSERT(!assertOnFatalError);
    DBG_LEAVE();
    return;
  }

#if DBG
    DbgPrint("RASPPTPE: EchoRequestTimeout %d cc: # %d s %d a %8.8x\n",
    controlChannelDescriptorp->echoIdentNumber,
    controlChannelDescriptorp->controlChannelIndex,
    controlChannelDescriptorp->controlChannelState,
    controlChannelDescriptorp->remoteNetConfigEntry.
    remoteControlChannelAddress.sin_addr.s_addr);
#endif

  /*
   * The timer waiting for a PPTP_ECHO_RESPONSE has expired before the
   * message being waited for has been received.  So initate shutdown
   * of the control channel.  For now we don't do this and will rely
   * on TCP to time out the control channel assuming it could never
   * send the PPTP_ECHO_REPLY message reliably
   */

	PptpNetControlDisconnectInd(controlChannelDescriptorp->controlChannelHandle);

	PptpNetControlDisconnectReq(controlChannelDescriptorp->controlChannelHandle, 1);

  DBG_LEAVE();

}

STATIC VOID PptpTerminateInactivityIdleProcessing(
	PptpControlChannelDescriptor	*controlChannelDescriptorp
	)
{
  DBG_FUNC("PptpTerminateInactivityIdleProcessing", PPTP_PROTOCOL);

  DBG_ENTER();

  /*
   * If there is an outstanding control channel inactivity timer scheduled
   * then cancel it
   */

  if (controlChannelDescriptorp->inactivityTimer.timerScheduled) {
    PptpUnscheduleInternalTimer(&controlChannelDescriptorp->inactivityTimer);
  }

  /*
   * Change the control channel inactivity state
   */

  controlChannelDescriptorp->controlChannelInactivityState =
    PPTP_CONTROL_CHANNEL_INACTIVITY_NO_STATE;

  DBG_LEAVE();

}

PptpReturnCode PptpScheduleInternalTimer(
	PptpCallbackFunction	callbackFunction,
	PptpCallbackArgument	callbackArgument,
	PptpMsec		milliSeconds,
	PptpInternalTimer	*timer
	)
{
	PptpMsec currentTimeMsec;

	DBG_FUNC("PptpScheduleInternalTimer", PPTP_PROTOCOL);

	DBG_ENTER();

	/*
	 * Validate the input arguments
	 */

	ASSERT(callbackFunction);

	/*
	 * Verify that the timer has not already been scheduled
	 */

	if (timer->timerScheduled) {
		ASSERT(!assertOnFatalError);
		DBG_LEAVE();
		return PPTP_FAILURE;
	}

	/*
	 * Set the callback fields in the internal timer structure
	 */

	PptpGetSystemTimeReq(&currentTimeMsec);
	timer->timerScheduled = 1;
	timer->callbackFunction = callbackFunction;
	timer->callbackArgument = callbackArgument;
	timer->expirationTimeMsec = currentTimeMsec + milliSeconds;

	/*
	 * Insert the timer into the internal timer list by checking
	 * the following order of conditions:
	 *
	 *   - see if there is anything in the list
	 *   - see if the timer goes at the front of the list
	 *   - insert the timer into the list by searching from the back
	 *     towards the front of the list
	 */

	if (internalTimerHead == NULL) {
		ASSERT(internalTimerTail == NULL);
		internalTimerHead = internalTimerTail = timer;
		timer->prev = timer->next = NULL;

		/*
		 * Since this were no entries in the timer list an external timer
		 * was not scheduled so schedule a new external timer to expire for
		 * the first entry in the list
		 */

//		ASSERT(externalTimerId == PPTP_TIMER_NOT_IN_USE);
//		if (PptpCreateExternalTimerReq(PptpExternalTimerCallback,
//			internalTimerHead,
//			internalTimerHead->expirationTimeMsec - currentTimeMsec,
//			250,
//			&externalTimerId) != PPTP_SUCCESS) {
//
//			ASSERT(!assertOnFatalError);
//			DBG_LEAVE();
//			return PPTP_FAILURE;
//		}
	} else if (timer->expirationTimeMsec < internalTimerHead->expirationTimeMsec) {
		timer->prev = NULL;
		timer->next = internalTimerHead;
		internalTimerHead->prev = timer;
		internalTimerHead = timer;

		/*
		 * The new timer is younger than the internal timer head so
		 * cancel the external timer scheduled for the internal timer head
		 * and schedule a new one for the new timer
		 */

//		ASSERT(externalTimerId != PPTP_TIMER_NOT_IN_USE);
//		if (PptpDeleteExternalTimerReq(externalTimerId) != PPTP_SUCCESS) {
//			ASSERT(!assertOnFatalError);
//		}
//
//		externalTimerId = PPTP_TIMER_NOT_IN_USE;

//		if (PptpCreateExternalTimerReq(PptpExternalTimerCallback,
//			internalTimerHead,
//			internalTimerHead->expirationTimeMsec - currentTimeMsec,
//			&externalTimerId) != PPTP_SUCCESS) {
//			ASSERT(!assertOnFatalError);
//			DBG_LEAVE();
//			return PPTP_FAILURE;
//		}
	} else {
		PptpInternalTimer *currentInternalTimer = internalTimerTail;
		while (1) {
			if (timer->expirationTimeMsec >=
				currentInternalTimer->expirationTimeMsec) {
				timer->prev = currentInternalTimer;
				timer->next = currentInternalTimer->next;
				if (currentInternalTimer->next) {
					currentInternalTimer->next->prev = timer;
				} else {
					internalTimerTail = timer;
				}
				currentInternalTimer->next = timer;
				break;
			}
			currentInternalTimer = currentInternalTimer->prev;
			ASSERT(currentInternalTimer);
		}
	}

	DBG_LEAVE();
	return PPTP_SUCCESS;
}


PptpReturnCode PptpUnscheduleInternalTimer(
	PptpInternalTimer	*timer
	)
{
	DBG_FUNC("PptpUnscheduleInternalTimer", PPTP_PROTOCOL);

    DBG_ENTER();

	/*
	 * Verify that the timer being unscheduled is currently scheduled
	 */

	if (!timer->timerScheduled) {
		ASSERT(!assertOnFatalError);
		DBG_LEAVE();
		return PPTP_FAILURE;
	}

	timer->timerScheduled = 0;

	/*
	 * If the internal timer was at the head of the list then unschedule
	 * the external timer and reschedule it if there is another entry in
	 * the list
	 */

//	if (timer == internalTimerHead) {

//		externalTimerId = PPTP_TIMER_NOT_IN_USE;

//		ASSERT(externalTimerId != PPTP_TIMER_NOT_IN_USE);

//		if (timer->next) {
//			PptpMsec currentTimeMsec;
//          PptpGetSystemTimeReq(&currentTimeMsec);
//			if (PptpCreateExternalTimerReq(PptpExternalTimerCallback,
//				timer->next,
//				timer->next->expirationTimeMsec - currentTimeMsec,
//				&externalTimerId) != PPTP_SUCCESS) {
//				ASSERT(!assertOnFatalError);
//			}
//		} else {
//			if (PptpDeleteExternalTimerReq(externalTimerId) != PPTP_SUCCESS) {
//				ASSERT(!assertOnFatalError);
//			}
//			externalTimerId = PPTP_TIMER_NOT_IN_USE;
//		}
//	}

	/*
	 * Remove the timer from the internal timer list
	 */

	if (timer->prev) {
		ASSERT(timer != internalTimerHead);
		timer->prev->next = timer->next;
	} else {
		ASSERT(timer == internalTimerHead);
		internalTimerHead = timer->next;
	}

	if (timer->next) {
		ASSERT(timer != internalTimerTail);
		timer->next->prev = timer->prev;
	} else {
		ASSERT(timer == internalTimerTail);
		internalTimerTail = timer->prev;
	}

	ASSERT((!internalTimerHead && !internalTimerTail) ||
	       (internalTimerHead && internalTimerTail));

	DBG_LEAVE();

	return PPTP_SUCCESS;
}


STATIC VOID PptpExternalTimerCallback(
	PptpCallbackArgument		callbackArgument
	)
{
	PptpMsec currentTimeMsec;

	DBG_FUNC("PptpExternalTimerCallback", PPTP_PROTOCOL);

	DBG_ENTER();

	/*
	 * The external timer has fired so set the external timer id to
	 * PPTP_TIMER_NOT_IN_USE
	 */

	externalTimerId = PPTP_TIMER_NOT_IN_USE;

	/*
	 * Incremement the current timer counter
	 */

	PptpGetSystemTimeReq(&currentTimeMsec);

	/*
	 * Perform some ASSERTs on the timer entries
	 */

#if defined(PPTP_USE_STRONG_LOCKS)
	ASSERT(g_PptpLockStrong.lockHeld);
#endif /* defined(PPTP_USE_STRONG_LOCKS) */
//	ASSERT(internalTimerHead);
//	ASSERT(callbackArgument == internalTimerHead);
//	ASSERT(internalTimerHead->expirationTimeMsec <= currentTimeMsec);
#if DBG
//	if (internalTimerHead->expirationTimeMsec > currentTimeMsec) {
//		DbgPrint("expirationTime %d > currentTime %d\n",
//		internalTimerHead->expirationTimeMsec, currentTimeMsec);
//	}
#endif

  /*
   * Check to see if any timer entries have expired and execute
   * their callbacks
   */

	while (internalTimerHead) {
		PptpInternalTimer *currentInternalTimer = internalTimerHead;
		if (internalTimerHead->expirationTimeMsec <= currentTimeMsec + 250) {
			ASSERT(currentInternalTimer->timerScheduled);
			ASSERT(currentInternalTimer->callbackFunction);
			ASSERT(currentInternalTimer->prev == NULL);
			currentInternalTimer->timerScheduled = 0;
			if (currentInternalTimer->next) {
				ASSERT(currentInternalTimer != internalTimerTail);
				currentInternalTimer->next->prev = NULL;
			}else {
				ASSERT(currentInternalTimer = internalTimerTail);
				internalTimerTail = NULL;
			}
			internalTimerHead = currentInternalTimer->next;
			currentInternalTimer->next = NULL;
			(*currentInternalTimer->callbackFunction)
			(currentInternalTimer->callbackArgument);
		} else {
		  break;
		}
	}
	
	ASSERT((!internalTimerHead && !internalTimerTail) ||
	       (internalTimerHead && internalTimerTail));

	/*
	 * If there are any more entries in the internal timer table and a
	 * new timer has not been scheduled then schedule a timer to fire
	 */
//	if (internalTimerHead) {
		if (PptpCreateExternalTimerReq(PptpExternalTimerCallback,
			internalTimerHead,
//			internalTimerHead->expirationTimeMsec - currentTimeMsec,
			250,
			&externalTimerId) != PPTP_SUCCESS) {
			ASSERT(!assertOnFatalError);
		}
//	}

	DBG_LEAVE();
}


STATIC PptpReturnCode PptpAssignNewControlChannelDescriptor(
	PptpControlChannelDescriptor	**controlChannelDescriptorpp
	)
{
  PptpControlChannelNumber controlChannelNumber;

  DBG_FUNC("PptpAssignNewControlChannelDescriptor", PPTP_PACKET_TRACE);

  DBG_ENTER();

  for (controlChannelNumber = 0 ;
       controlChannelNumber < currentPptpConfig->numberLineDevices ;
       controlChannelNumber++) {
    if (controlChannelDescriptorTable[controlChannelNumber].
	controlChannelState == PPTP_CONTROL_CHANNEL_NO_STATE) {
      controlChannelDescriptorTable[controlChannelNumber].controlChannelState =
	PPTP_CONTROL_CHANNEL_ALLOCATED;
      *controlChannelDescriptorpp =
	&controlChannelDescriptorTable[controlChannelNumber];
      DBG_LEAVE();
      return PPTP_SUCCESS;
    }
  }

  DBG_LEAVE();
  return PPTP_FAILURE;

}


STATIC VOID PptpTraceControlPacket(
	PptpControlPacket		*inputPacketp,
	BOOL					inputPacket
	)
{
  DBG_FUNC("PptpTraceControlPacket", PPTP_PACKET_TRACE);

  DBG_ENTER();

  DBG_TRACE("PptpTraceControlPacket: %s ",
	    (inputPacket ? "received" : "sending"));

  switch (ntohs(inputPacketp->header.messageType)) {
  case PPTP_START_SESSION_REQUEST: {
    PptpStartSessionRequestMsg *pMsg = &inputPacketp->startSessionRequest;
    if (ntohs(inputPacketp->header.packetLength) != sizeof(*pMsg)) {
      goto BadPacketLength;
    }
    PPTP_CONTROL_PACKET_LOG(("PPTP_START_SESSION_REQUEST\n"));
    PPTP_CONTROL_PACKET_LOG(("  protocolVersion = %d\n",
			     ntohs(pMsg->protocolVersion)));
    PPTP_CONTROL_PACKET_LOG(("  framingCapability = %d\n",
			     ntohl(pMsg->fepInfo.framingCapability)));
    PPTP_CONTROL_PACKET_LOG(("  bearerCapability = %d\n",
			     ntohl(pMsg->fepInfo.bearerCapability)));
    PPTP_CONTROL_PACKET_LOG(("  maxChannels = %d\n",
			     ntohs(pMsg->fepInfo.maxChannels)));
    PPTP_CONTROL_PACKET_LOG(("  firmwareRevision = %d\n",
			     ntohs(pMsg->fepInfo.firmwareRevision)));
    PPTP_CONTROL_PACKET_LOG(("  hostName = %s\n",
			     pMsg->fepInfo.hostName));
    PPTP_CONTROL_PACKET_LOG(("  vendorString = %s\n",
			     pMsg->fepInfo.vendorString));
    break;
  }
  case PPTP_START_SESSION_REPLY: {
    PptpStartSessionReplyMsg *pMsg = &inputPacketp->startSessionReply;
    if (ntohs(inputPacketp->header.packetLength) != sizeof(*pMsg)) {
      goto BadPacketLength;
    }
    PPTP_CONTROL_PACKET_LOG(("PPTP_START_SESSION_REPLY\n"));
    PPTP_CONTROL_PACKET_LOG(("  protocolVersion = %d\n",
			     ntohs(pMsg->protocolVersion)));
    PPTP_CONTROL_PACKET_LOG(("  resultCode = %d\n",
			     pMsg->resultCode));
    PPTP_CONTROL_PACKET_LOG(("  generalErrorCode = %d\n",
			     pMsg->generalErrorCode));
    PPTP_CONTROL_PACKET_LOG(("  framingCapability = %d\n",
			     ntohl(pMsg->fepInfo.framingCapability)));
    PPTP_CONTROL_PACKET_LOG(("  bearerCapability = %d\n",
			     ntohl(pMsg->fepInfo.bearerCapability)));
    PPTP_CONTROL_PACKET_LOG(("  maxChannels = %d\n",
			     ntohs(pMsg->fepInfo.maxChannels)));
    PPTP_CONTROL_PACKET_LOG(("  firmwareRevision = %d\n",
			     ntohs(pMsg->fepInfo.firmwareRevision)));
    PPTP_CONTROL_PACKET_LOG(("  hostName = %s\n",
			     pMsg->fepInfo.hostName));
    PPTP_CONTROL_PACKET_LOG(("  vendorString = %s\n",
			     pMsg->fepInfo.vendorString));
    break;
  }
  case PPTP_STOP_SESSION_REQUEST: {
    PptpStopSessionRequestMsg *pMsg = &inputPacketp->stopSessionRequest;
    if (ntohs(inputPacketp->header.packetLength) != sizeof(*pMsg)) {
      goto BadPacketLength;
    }
    PPTP_CONTROL_PACKET_LOG(("PPTP_STOP_SESSION_REQUEST\n"));
    PPTP_CONTROL_PACKET_LOG(("  reason = %d\n",
			     pMsg->reason));
    break;
  }
  case PPTP_STOP_SESSION_REPLY: {
    PptpStopSessionReplyMsg *pMsg = &inputPacketp->stopSessionReply;
    if (ntohs(inputPacketp->header.packetLength) != sizeof(*pMsg)) {
      goto BadPacketLength;
    }
    PPTP_CONTROL_PACKET_LOG(("PPTP_STOP_SESSION_REPLY\n"));
    PPTP_CONTROL_PACKET_LOG(("  resultCode = %d\n",
			     pMsg->resultCode));
    PPTP_CONTROL_PACKET_LOG(("  generalErrorCode = %d\n",
			     pMsg->generalErrorCode));
    break;
  }
  case PPTP_ECHO_REQUEST: {
    PptpEchoRequestMsg *pMsg = &inputPacketp->echoRequest;
    if (ntohs(inputPacketp->header.packetLength) != sizeof(*pMsg)) {
      goto BadPacketLength;
    }
    PPTP_CONTROL_PACKET_LOG(("PPTP_ECHO_REQUEST\n"));
    PPTP_CONTROL_PACKET_LOG(("  identNumber = %d\n",
			     ntohl(pMsg->identNumber)));
    break;
  }
  case PPTP_ECHO_REPLY: {
    PptpEchoReplyMsg *pMsg = &inputPacketp->echoReply;
    if (ntohs(inputPacketp->header.packetLength) != sizeof(*pMsg)) {
      goto BadPacketLength;
    }
    PPTP_CONTROL_PACKET_LOG(("PPTP_ECHO_REPLY\n"));
    PPTP_CONTROL_PACKET_LOG(("  identNumber = %d\n",
			     ntohl(pMsg->identNumber)));
    PPTP_CONTROL_PACKET_LOG(("  resultCode = %d\n",
			     pMsg->resultCode));
    PPTP_CONTROL_PACKET_LOG(("  generalErrorCode = %d\n",
			     pMsg->generalErrorCode));
    break;
  }
  case PPTP_OUT_CALL_REQUEST: {
    PptpOutCallRequestMsg *pMsg = &inputPacketp->outCallRequest;
    if (ntohs(inputPacketp->header.packetLength) != sizeof(*pMsg)) {
      goto BadPacketLength;
    }
    PPTP_CONTROL_PACKET_LOG(("PPTP_OUT_CALL_REQUEST\n"));
    PPTP_CONTROL_PACKET_LOG(("  callID = %d\n",
			     ntohs(pMsg->callID)));
    PPTP_CONTROL_PACKET_LOG(("  callSerialNumber = %d\n",
			     ntohs(pMsg->callSerialNumber)));
    PPTP_CONTROL_PACKET_LOG(("  minBPS = %d\n",
			     ntohl(pMsg->minBPS)));
    PPTP_CONTROL_PACKET_LOG(("  maxBPS = %d\n",
			     ntohl(pMsg->maxBPS)));
    PPTP_CONTROL_PACKET_LOG(("  bearerType = %d\n",
			     ntohl(pMsg->bearerType)));
    PPTP_CONTROL_PACKET_LOG(("  framingType = %d\n",
			     ntohl(pMsg->framingType)));
    PPTP_CONTROL_PACKET_LOG(("  packetWindow = %d\n",
			     ntohs(pMsg->packetWindow)));
    PPTP_CONTROL_PACKET_LOG(("  packetProcDelay = %d\n",
			     ntohs(pMsg->packetProcDelay)));
    PPTP_CONTROL_PACKET_LOG(("  phoneNumberLength = %d\n",
			     ntohs(pMsg->phoneNumberLength)));
    PPTP_CONTROL_PACKET_LOG(("  phoneNumber = %s\n",
			     pMsg->phoneNumber));
    PPTP_CONTROL_PACKET_LOG(("  subAddress = %s\n",
			     pMsg->subAddress));
    break;
  }
  case PPTP_OUT_CALL_REPLY: {
    PptpOutCallReplyMsg *pMsg = &inputPacketp->outCallReply;
    if (ntohs(inputPacketp->header.packetLength) != sizeof(*pMsg)) {
      goto BadPacketLength;
    }
    PPTP_CONTROL_PACKET_LOG(("PPTP_OUT_CALL_REPLY\n"));
    PPTP_CONTROL_PACKET_LOG(("  callID = %d\n",
			     ntohs(pMsg->callID)));
    PPTP_CONTROL_PACKET_LOG(("  peerCallID = %d\n",
			     ntohs(pMsg->peerCallID)));
    PPTP_CONTROL_PACKET_LOG(("  resultCode = %d\n",
			     pMsg->resultCode));
    PPTP_CONTROL_PACKET_LOG(("  generalErrorCode = %d\n",
			     pMsg->generalErrorCode));
    PPTP_CONTROL_PACKET_LOG(("  causeCode = %d\n",
			     ntohs(pMsg->causeCode)));
    PPTP_CONTROL_PACKET_LOG(("  connectSpeed = %d\n",
			     ntohl(pMsg->connectSpeed)));
    PPTP_CONTROL_PACKET_LOG(("  packetWindow = %d\n",
			     ntohs(pMsg->packetWindow)));
    PPTP_CONTROL_PACKET_LOG(("  packetProcDelay = %d\n",
			     ntohs(pMsg->packetProcDelay)));
    PPTP_CONTROL_PACKET_LOG(("  physChannelID = %d\n",
			     ntohl(pMsg->physChannelID)));
    break;
  }
  case PPTP_IN_CALL_REQUEST: {
    PptpInCallRequestMsg *pMsg = &inputPacketp->inCallRequest;
    if (ntohs(inputPacketp->header.packetLength) != sizeof(*pMsg)) {
      goto BadPacketLength;
    }
    PPTP_CONTROL_PACKET_LOG(("PPTP_IN_CALL_REQUEST\n"));
    PPTP_CONTROL_PACKET_LOG(("  callID = %d\n",
			     ntohs(pMsg->callID)));
    PPTP_CONTROL_PACKET_LOG(("  callSerialNumber = %d\n",
			     ntohs(pMsg->callSerialNumber)));
    PPTP_CONTROL_PACKET_LOG(("  bearerType = %d\n",
			     ntohl(pMsg->bearerType)));
    PPTP_CONTROL_PACKET_LOG(("  physChannelID = %d\n",
			     ntohl(pMsg->physChannelID)));
    PPTP_CONTROL_PACKET_LOG(("  dialedNumberLength = %d\n",
			     ntohs(pMsg->dialedNumberLength)));
    PPTP_CONTROL_PACKET_LOG(("  dialingNumberLength = %d\n",
			     ntohs(pMsg->dialingNumberLength)));
    PPTP_CONTROL_PACKET_LOG(("  dialedNumber = %s\n",
			     pMsg->dialedNumber));
    PPTP_CONTROL_PACKET_LOG(("  dialingNumber = %s\n",
			     pMsg->dialingNumber));
    PPTP_CONTROL_PACKET_LOG(("  subAddress = %s\n",
			     pMsg->subAddress));
    break;
  }
  case PPTP_IN_CALL_REPLY: {
    PptpInCallReplyMsg *pMsg = &inputPacketp->inCallReply;
    if (ntohs(inputPacketp->header.packetLength) != sizeof(*pMsg)) {
      goto BadPacketLength;
    }
    PPTP_CONTROL_PACKET_LOG(("PPTP_IN_CALL_REPLY\n"));
    PPTP_CONTROL_PACKET_LOG(("  callID = %d\n",
			     ntohs(pMsg->callID)));
    PPTP_CONTROL_PACKET_LOG(("  peerCallID = %d\n",
			     ntohs(pMsg->peerCallID)));
    PPTP_CONTROL_PACKET_LOG(("  resultCode = %d\n",
			     pMsg->resultCode));
    PPTP_CONTROL_PACKET_LOG(("  generalErrorCode = %d\n",
			     pMsg->generalErrorCode));
    PPTP_CONTROL_PACKET_LOG(("  packetWindow = %d\n",
			     ntohs(pMsg->packetWindow)));
    PPTP_CONTROL_PACKET_LOG(("  packetProcDelay = %d\n",
			     ntohs(pMsg->packetProcDelay)));
    break;
  }
  case PPTP_IN_CALL_CONNECTED: {
    PptpInCallConnectedMsg *pMsg = &inputPacketp->inCallConnected;
    if (ntohs(inputPacketp->header.packetLength) != sizeof(*pMsg)) {
      goto BadPacketLength;
    }
    PPTP_CONTROL_PACKET_LOG(("PPTP_IN_CALL_CONNECTED\n"));
    PPTP_CONTROL_PACKET_LOG(("  callID = %d\n",
			     ntohs(pMsg->callID)));
    PPTP_CONTROL_PACKET_LOG(("  peerCallID = %d\n",
			     ntohs(pMsg->peerCallID)));
    PPTP_CONTROL_PACKET_LOG(("  connectSpeed = %d\n",
			     ntohl(pMsg->connectSpeed)));
    PPTP_CONTROL_PACKET_LOG(("  packetWindow = %d\n",
			     ntohs(pMsg->packetWindow)));
    PPTP_CONTROL_PACKET_LOG(("  packetProcDelay = %d\n",
			     ntohs(pMsg->packetProcDelay)));
    PPTP_CONTROL_PACKET_LOG(("  framingType = %d\n",
			     ntohl(pMsg->framingType)));
    break;
  }
 case PPTP_CALL_CLEAR_REQUEST: {
    PptpCallClearRequestMsg *pMsg = &inputPacketp->callClearRequest;
    if (ntohs(inputPacketp->header.packetLength) != sizeof(*pMsg)) {
      goto BadPacketLength;
    }
    PPTP_CONTROL_PACKET_LOG(("PPTP_CALL_CLEAR_REQUEST\n"));
    PPTP_CONTROL_PACKET_LOG(("  callID = %d\n",
			     ntohs(pMsg->callID)));
    break;
  }
  case PPTP_CALL_DISCONNECT_NOTIFY: {
    PptpCallDisconnectNotifyMsg *pMsg = &inputPacketp->callDisconnectNotify;
    if (ntohs(inputPacketp->header.packetLength) != sizeof(*pMsg)) {
      goto BadPacketLength;
    }
    PPTP_CONTROL_PACKET_LOG(("PPTP_CALL_DISCONNECT_NOTIFY\n"));
    PPTP_CONTROL_PACKET_LOG(("  callID = %d\n",
			     ntohs(pMsg->callID)));
    PPTP_CONTROL_PACKET_LOG(("  resultCode = %d\n",
			     pMsg->resultCode));
    PPTP_CONTROL_PACKET_LOG(("  generalErrorCode = %d\n",
			     pMsg->generalErrorCode));
    PPTP_CONTROL_PACKET_LOG(("  causeCode = %d\n",
			     ntohs(pMsg->causeCode)));
    PPTP_CONTROL_PACKET_LOG(("  callStatistics = %s\n",
			     pMsg->callStatistics));
    break;
  }
  case PPTP_WAN_ERROR_NOTIFY: {
    PptpWanErrorNotifyMsg *pMsg = &inputPacketp->wanErrorNotify;
    if (ntohs(inputPacketp->header.packetLength) != sizeof(*pMsg)) {
      goto BadPacketLength;
    }
    PPTP_CONTROL_PACKET_LOG(("PPTP_WAN_ERROR_NOTIFY\n"));
    PPTP_CONTROL_PACKET_LOG(("  peerCallID = %d\n",
			     ntohs(pMsg->peerCallID)));
    PPTP_CONTROL_PACKET_LOG(("  crcErrors = %d\n",
			     ntohl(pMsg->crcErrors)));
    PPTP_CONTROL_PACKET_LOG(("  framingErrors = %d\n",
			     ntohl(pMsg->framingErrors)));
    PPTP_CONTROL_PACKET_LOG(("  hardwareOverRuns = %d\n",
			     ntohl(pMsg->hardwareOverRuns)));
    PPTP_CONTROL_PACKET_LOG(("  bufferOverRuns = %d\n",
			     ntohl(pMsg->bufferOverRuns)));
    PPTP_CONTROL_PACKET_LOG(("  timeoutErrors = %d\n",
			     ntohl(pMsg->timeoutErrors)));
    PPTP_CONTROL_PACKET_LOG(("  alignmentErrors = %d\n",
			     ntohl(pMsg->alignmentErrors)));
    break;
  }
  case PPTP_SET_LINK_INFO: {
    PptpSetLinkInfoMsg *pMsg = &inputPacketp->setLinkInfo;
    if (ntohs(inputPacketp->header.packetLength) != sizeof(*pMsg)) {
      goto BadPacketLength;
    }
    PPTP_CONTROL_PACKET_LOG(("PPTP_SET_LINK_INFO\n"));
    PPTP_CONTROL_PACKET_LOG(("  peerCallID = %d\n",
			     ntohs(pMsg->peerCallID)));
    PPTP_CONTROL_PACKET_LOG(("  sendAccm = %Xh\n",
			     ntohl(pMsg->sendAccm)));
    PPTP_CONTROL_PACKET_LOG(("  recvAccm = %Xh\n",
			     ntohl(pMsg->recvAccm)));
    break;
  }
  default:
	  PPTP_CONTROL_PACKET_LOG(("unrecognized messageType = %d\n",
				   inputPacketp->header.messageType));
    break;
  }
  DBG_LEAVE();
  return;

  BadPacketLength:
  PPTP_CONTROL_PACKET_LOG(("messageType = %d with bad packetLength = %d\n",
			   inputPacketp->header.messageType,
			   inputPacketp->header.packetLength));
  DBG_LEAVE();

}


STATIC VOID PptpTraceDataPacket(
	PptpCallLineNumber	callLineNumber,
	PptpSeqNumber		seqNumber,
	PptpAckNumber		ackNumber,
	UCHAR			*data,
	ULONG			dataLength,
	BOOL			inputPacket
	)
{
  DBG_FUNC("PptpTraceDataPacket", PPTP_PACKET_TRACE);

  DBG_ENTER();

  PPTP_DATA_PACKET_HEADER_LOG(("PptpTraceDataPacket: %s PPTP data packet\n",
			       (inputPacket ? "received" : "sending")));

  PPTP_DATA_PACKET_HEADER_LOG(("  callLineNumber = %d\n", callLineNumber));
  PPTP_DATA_PACKET_HEADER_LOG(("  seqNumber = %d\n", seqNumber));
  PPTP_DATA_PACKET_HEADER_LOG(("  ackNumber = %d\n", ackNumber));
  PPTP_DATA_PACKET_HEADER_LOG(("  dataLength = %d\n", dataLength));

  PPTP_DATA_PACKET_DATA_LOG(data, dataLength);

  DBG_LEAVE();
}

STATIC UCHAR *PptpAddressToString(
	PptpSockaddrIn			*address,
	UCHAR				*buffer
	)
{
  LONG position;

  buffer[0] = '\0';

  for (position = 3 ; position >= 0 ; position--) {
    UCHAR value = (UCHAR) (address->sin_addr.s_addr >> (position * 8));
    if (value > 0) {
      ULONG base = 100;
      while (base > 0) {
	if ((value / base) > 0) {
	  UCHAR number[2];
	  number[0] = (UCHAR) (value / base + '0');
	  number[1] = '\0';
	  strcat(buffer, number);
	  value = (UCHAR) (value % base);
	}
	base /= 10;
      }
    }
    else {
      strcat(buffer, "0");
    }
    if (position > 0) {
      strcat(buffer, ".");
    }
  }

  return buffer;

}

