/****************************************************************************
 *  pptpcomm.h    : $Revision:   1.31  $
 *  Modified      : $Date:   May 13 1996 14:48:36  $
 *  By            : $Author:   tmortsol  $, U.S. Robotics Access Corp.
 *****************************************************************************
 *          Copyright (c) $Date:   May 13 1996 14:48:36  $,                  *
 *            U.S. Robotics Access Corp., Microsoft Corp.                    *
 *                       All Rights Reserved                                 *
 *****************************************************************************
 *                          "PPTP MODULE"                                    *
 * This software may be used to implement the POINT to POINT TUNNELING       *
 * PROTOCOL as defined by the PPTP forum.  The PPTP forum consists of:       *
 * U.S. Robotics, Microsoft, Ascend Communications, 3Com/Primary Access,     *
 * Telematics.                                                               *
 *****************************************************************************
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain this entire header
 *    including the above copyright notice, this list of conditions and
 *    the following disclaimer.
 * 2. The name of U.S. Robotics may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission of U.S. Robotics.
 *
 * THIS SOFTWARE IS PROVIDED BY U.S. ROBOTICS ACESS CORP. ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL U.S. ROBOTICS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/
/*END_HEADER*/

/*
 * pptpcomm.h
 *
 * This file contains the common definitions used by the PPTP protocol.
 * It contains the following types of definitions:
 *
 *   - object typedefs
 *   - common #defines such as limits
 *
 * This header file should be included before any other PPTP header file.
 *
 * Author: Tim Mortsolf, U.S. Robotics Access Corp.
 */

#if !defined(_pptpcomm_h)
#define _pptpcomm_h

/*
 * Typedefs for the standard types defined in the PPTP protocol
 * header file.  Define PPTP_DATA_MESSAGE_TYPES to indicate that these
 * types have been defined so they do not get redefined.
 */

#if !defined(PPTP_DATA_MESSAGE_TYPES)
#define PPTP_DATA_MESSAGE_TYPES

typedef unsigned char Byte;
typedef unsigned short Word;
typedef unsigned long LongWord;

#endif /* !defined(PPTP_DATA_MESSAGE_TYPES) */

/*
 * Include "pptppro.h" to get definitions from the protocol header
 * file that are referenced in this file
 */

#include "pptppro.h"

/*
 * Based upon the definition of the PPTP_TARGET define, include the
 * appropriate common header file that defines target specific definitions.
 * The following variables of PPTP_TARGET are currently recognized:
 *
 *   PPTP_TARGET_NDIS_MINIPORT_DRIVER
 *     Compile the code as an NDIS miniport driver by including
 *     pcomndis.h
 *
 *   PPTP_TARGET_WIN32
 *     Compile the code as a Win32 user level application by
 *     including pcomw32.h
 *
 *   PPTP_TARGET_FEP
 *     Compile the code as a FEP by including pcomfep.h
 *
 *   PPTP_TARGET_STANDALONE
 *     Compile the protocol engine as a standalone object by including pptpsa.h
 */

#define PPTP_TARGET_NDIS_MINIPORT_DRIVER 0
#define PPTP_TARGET_WIN32 1
#define PPTP_TARGET_FEP 2
#define PPTP_TARGET_STANDALONE 3

#if !defined(PPTP_TARGET)

/*
 * The value of PPTP_TARGET is not defined so try to guess what target is
 * being built.
 *
 *   - if NDIS_MINIPORT_DRIVER is defined then assume the source file
 *     is compiling as a NT driver
 *   - else if USR_PPTP is defined then assume the source file
 *     is compiling as a FEP target
 *   - else assume the source file is compiling as a Win32 user level app
 */

#if defined(NDIS_MINIPORT_DRIVER)
#define PPTP_TARGET PPTP_TARGET_NDIS_MINIPORT_DRIVER
#elif defined(USR_PPTP)
#define PPTP_TARGET PPPT_TARGET_FEP
#else
#define PPTP_TARGET PPTP_TARGET_WIN32
#endif /* defined(NDIS_MINIPORT_DRIVER) */

#endif /* !defined(PPTP_TARGET) */

#if (PPTP_TARGET == PPTP_TARGET_NDIS_MINIPORT_DRIVER)
#include "pcomndis.h"
#elif (PPTP_TARGET == PPTP_TARGET_WIN32)
#include "pcomw32.h"
#elif (PPTP_TARGET == PPTP_TARGET_FEP)
#include "pcomfep.h"
#elif (PPTP_TARGET == PPTP_TARGET_STANDALONE)
#include "pcomsa.h"
#else
ERROR - PPTP_TARGET definition is not valid
#endif /* PPTP_TARGET */

/*
 * #defines used for standard lengths not defined in the PPTP protocol
 * header file
 */

#define PPTP_MAX_ADDRESS_LENGTH (80)

/*
 * #defines for general quantitites of entities
 */

#define PPTP_MAX_CALL_LINE_NUMBER (256)
#define PPTP_MAX_CONTROL_CHANNEL_NUMBER (PPTP_MAX_CALL_LINE_NUMBER)
#define PPTP_MAX_CONTROL_CHANNEL_SEEK_ADDRESS_ENTRIES (8)
#define PPTP_MAX_CONTROL_CHANNEL_SEEK_NUMBER (4)
#define PPTP_MAX_PACKET_WINDOW (256)

/*
 * #defines for unspecified values
 */

#define PPTP_UNUSED_CALL_LINE_NUMBER ((PptpCallLineNumber) (-1))
#define PPTP_ENTRY_NOT_IN_USE ((ULONG) -1)
#define PPTP_TIMER_NOT_IN_USE ((PptpExternalTimerId) -1)

/*
 * Typedefs for standard types used internally by our PPTP implementation.
 */

typedef int BOOL;
typedef unsigned char UCHAR;
typedef unsigned short USHORT;
typedef unsigned long ULONG;
typedef void * PVOID;

/*
 * Typedefs for types defined specifically for use by the PPTP protocol
 */

typedef USHORT PptpCallLineNumber;
typedef PVOID PptpCallDeviceNumber;
typedef USHORT PptpCallSerialNumber;
typedef ULONG PptpPhysChannelID;
typedef ULONG PptpControlChannelNumber;
typedef ULONG PptpNetConfigIndex;
typedef ULONG PptpControlChannelIndex;
typedef ULONG PptpControlChannelSeekIndex;
typedef ULONG PptpAddressListIndex;
typedef USHORT PptpPortNumber;
typedef USHORT PptpProtocolNumber;
typedef ULONG PptpEndpointHandle;
typedef USHORT PptpPacketWindow;
typedef USHORT PptpPacketProcDelay;
typedef ULONG PptpSeqNumber;
typedef ULONG PptpAckNumber;
typedef PVOID PptpExternalTimerId;
typedef LONG PptpSec;
typedef LONG PptpMsec;
typedef USHORT PptpCauseCode;
typedef ULONG PptpBPS;
typedef struct sockaddr_in PptpSockaddrIn;
typedef ULONG PptpPeerIndex;
typedef ULONG PptpAccmMap;
typedef ULONG PptpEchoIdentNumber;
typedef PVOID PptpCallbackArgument;
typedef VOID (*PptpCallbackFunction)(PptpCallbackArgument callbackArgument);

typedef enum {
  PPTP_SUCCESS = 0,
  PPTP_FAILURE
} PptpReturnCode;

typedef enum {
  PPTP_DATA_CHANNEL_FLOW_CONTROL_DISABLED,
  PPTP_DATA_CHANNEL_FLOW_CONTROL_STATIC_TIMEOUT,
  PPTP_DATA_CHANNEL_FLOW_CONTROL_SLIDING_WINDOW
} PptpDataChannelFlowControlType;

typedef struct {
  USHORT addressLength;
  UCHAR	address[PPTP_MAX_ADDRESS_LENGTH];
} PptpCallAddress;
	
typedef struct {
  PptpPortNumber localControlChannelPortNumber;
  PptpProtocolNumber localProtocolNumber;
  PptpFepInfo fepInfo;
} PptpLocalNetConfig;

typedef struct {
  BOOL entryInUse;
  PptpSockaddrIn remoteControlChannelAddress;
  PptpProtocolNumber remoteProtocolNumber;
} PptpRemoteNetConfigEntry;

typedef struct {
  PptpRemoteNetConfigEntry
    addressTable[PPTP_MAX_CONTROL_CHANNEL_SEEK_ADDRESS_ENTRIES];
} PptpAddressList;

typedef struct {
  PptpLocalNetConfig localNetConfig;
  PptpAddressList defaultRemoteAddressList;
  BOOL tunnelGreOverUdp;
} PptpNetConfig;

typedef struct {
  PptpNetConfig netConfig;
  ULONG numberLineDevices;
  BOOL authenticateIncomingControlChannelAddress;
  PptpDataChannelFlowControlType dataChannelFlowControlType;
  PptpSec dataChannelFlowControlStaticTimeoutSeconds;
  PptpPacketWindow rxPacketWindow;
  PptpSec inactivityIdleSeconds;
  PptpSec echoReplyTimeoutSeconds;
  BOOL AssertOnProtocolError;
  BOOL AssertOnFatalError;
  BOOL AssertOnMemoryAllocationFailure;
} PptpConfig;

/*
 * Typedefs for definitions that should come from pptppro.h
 */

typedef ULONG PptpWanErrorCode;

/*
 * Define assert and log definitions
 */

#define PptpLogEvent PptpPrintf
#define PptpLogError PptpPrintf

/*
 * Define macros to lock criticial sections of code
 * There are three different methods of locks that can be compiled
 * into the driver, these methods are differentiated by the
 * definition of the PPTP_USE_STRONG_LOCKS macro:
 *
 *   - Strong locks are enabled if PPTP_USE_STRONG_LOCKS is defined
 *     Strong locks are simple entry/exit locks placed in all PPTP asynchronous
 *     routines that prevent any other PPTP asynchronous routines from
 *     running.  These locks are not as efficient and strategically placed
 *     as standard locks and are meant to be used for debugging to see if
 *     a problem is due to a race condition.  If a problem still persists
 *     when strong locks are enabled then it is probably not due to a
 *     resource contention problem
 *
 *   ** NOTE ** standard locks are not yet completely implemented
 *
 *   - Standard locks are enabled if PPTP_USE_STANDARD_LOCKS is defined
 *     Standard locks are strategically placed throughout the code
 *     to not hold onto spinlocks as lock as a strong lock will.  Standard
 *     locks also use different locking variables to lock on.  The strategy
 *     behind the standard locks is:
 *
 *       o Serialize all call control events so they appear to run as a
 *         single thread.  This makes the synchronization of the call
 *         control code a lot simpler since simple exit/entry locks
 *         will be used for all call control.  The lock used to serialize
 *         all call control events is PPTP_LOCK_GLOBAL()/PPTP_UNLOCK_GLOBAL()
 *         Locking of the control channels and call lines is done to prevent
 *         any synchronization problems of call control and data events
 *       o Put locks in the data path so multiple calls can handle data
 *         at the same simulataneously on a multi-processor machine.  This
 *         allows the protocol to take advantage of the additional performance
 *         offered by an SMP operating system
 *
 *   - No locks are enabled if PPTP_USE_NO_LOCKS is defined.  In this
 *     case all the locking macros evaluate to nothing
 */

#if defined(PPTP_USE_STRONG_LOCKS)

extern PptpLock g_PptpLockStrong;

#define PPTP_LOCK_STRONG() \
  PptpPrintf("PPTP_LOCK_STRONG()\n"); \
  PPTP_LOCK(g_PptpLockStrong);

#define PPTP_UNLOCK_STRONG() \
  PptpPrintf("PPTP_UNLOCK_STRONG()\n"); \
  PPTP_UNLOCK(g_PptpLockStrong);

#define PPTP_LOCK_GLOBAL()
#define PPTP_UNLOCK_GLOBAL()

#define PPTP_LOCK_CONTROL_CHANNEL(ControlChannelLock)
#define PPTP_UNLOCK_CONTROL_CHANNEL(ControlChannelLock)

#define PPTP_LOCK_CALL_LINE(CallLineLock)
#define PPTP_UNLOCK_CALL_LINE(CallLineLock)

#elif defined(PPTP_USE_STANDARD_LOCKS)

ERROR - standard locks not yet completely implemented

extern PptpLock g_PptpLockGlobal;

#define PPTP_LOCK_STRONG()
#define PPTP_UNLOCK_STRONG()

#define PPTP_LOCK_GLOBAL() \
  PptpPrintf("PPTP_LOCK_GLOBAL()\n"); \
  PPTP_LOCK(g_PptpLockGlobal);

#define PPTP_UNLOCK_GLOBAL() \
  PptpPrintf("PPTP_UNLOCK_GLOBAL()\n"); \
  PPTP_UNLOCK(g_PptpLockGlobal);

#define PPTP_LOCK_CONTROL_CHANNEL(ControlChannelLock) \
  PptpPrintf("PPTP_LOCK_CONTROL_CHANNEL()\n"); \
  PPTP_LOCK(ControlChannelLock);

#define PPTP_UNLOCK_CONTROL_CHANNEL(ControlChannelLock) \
  PptpPrintf("PPTP_UNLOCK_CONTROL_CHANNEL()\n"); \
  PPTP_UNLOCK(ControlChannelLock);

#define PPTP_LOCK_CALL_LINE(CallLineLock) \
  PptpPrintf("PPTP_LOCK_CALL_LINE()\n"); \
  PPTP_LOCK(CallLineLock);

#define PPTP_UNLOCK_CALL_LINE(CallLineLock) \
  PptpPrintf("PPTP_UNLOCK_CALL_LINE()\n"); \
  PPTP_UNLOCK(CallLineLock);

#elif defined(PPTP_USE_NO_LOCKS)

#define PPTP_LOCK_INIT()

#define PPTP_LOCK_STRONG()
#define PPTP_UNLOCK_STRONG()

#define PPTP_LOCK_GLOBAL()
#define PPTP_UNLOCK_GLOBAL()

#define PPTP_LOCK_CONTROL_CHANNEL(ControlChannelLock)
#define PPTP_UNLOCK_CONTROL_CHANNEL(ControlChannelLock)

#define PPTP_LOCK_CALL_LINE(CallLineLock)
#define PPTP_UNLOCK_CALL_LINE(CallLineLock)

#else

ERROR - not locking mechanism was defined

#endif /* defined(PPTP_USE_STRONG_LOCKS) */

#endif /* !defined(_pptpcomm_h) */
